# encoding: utf-8
# TO DO: unicode error
import sys
import os
import signal
import webbrowser
import platform
try:
    from PyQt4.QtGui import *
    from PyQt4.QtCore import *  #
except ImportError:
    sys.stderr.write(
        "Module PyQt4 is required for the graphical interface of PAIREF. "
        "Please install it or execute a PAIREF job from command-line.\n"
        "Tip: PyQt4 is installed in ccp4-python.\n"
        "Aborting.\n")
    sys.exit(1)


class MyWindow(QWidget):
    """PAIREF graphical interface (in PyQt4)"""
    def __init__(self):
        super(MyWindow, self).__init__()
        # Allow to term the app pressing Ctrl+C in console
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        # locale.setlocale(locale.LC_ALL, "en_US.UTF-8")  useless
        # QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))
        self.myInit()

    def myInit(self):
        """Specifications of widgets, their layout and connection"""
        self.setWindowTitle("PAIREF")
        logofilename = \
            str(os.path.dirname(__file__)) + os.path.sep + \
            'static' + os.path.sep + 'pairef_logo_64.png'
        self.setStyleSheet("""
QPushButton#required {
    font-weight: bold;
}

QLineEdit#required {
    background-color: lightyellow;
}

#indent {
    margin-left: 2em;
}

#italic {
    font-style: italic;
}
""")
        intValidatorPositive = QIntValidator()  # only real numbers
        intValidatorPositive.setRange(1, 999999)
        doubleValidator = QDoubleValidator()  # only real numbers
        # doubleValidatorPositive = QDoubleValidator()  # only real positive numbers
        # doubleValidatorPositive.setRange(0.000001, 999999)  # Not working well
        doubleValidatorPositive = doubleValidator  # Accepts also negative :-(
        alphanumValidator = QRegExpValidator(QRegExp("[A-Za-z0-9_-]+"))
        
        if os.path.isfile(logofilename):
            self.logoLabel = QLabel(self)
            logo = QPixmap(logofilename)
            self.logoLabel.setPixmap(logo)
            self.logoLabel.setAlignment(Qt.AlignCenter)
        self.header = QLabel("Automatic PAIRed REFinement protocol", 
                             self, objectName="italic")
        self.header.setAlignment(Qt.AlignCenter)

        resolutionRadio = QButtonGroup(self)
        self.step_numberRadio = QRadioButton("Define resolution step")
        resolutionRadio.addButton(self.step_numberRadio)
        self.step_numberRadio.setChecked(True)
        self.step_numberRadio.toggled.connect(self.resToggled)
        self.stepLabel = QLabel("Add high-resolution shells with a width of",
                                self, objectName="indent")
        self.stepLabel.setToolTip("in angstrom")
        self.step = QLineEdit("0.05")
        self.step.setValidator(doubleValidatorPositive)
        self.step.setToolTip("in angstrom")
        self.numberLabel = QLabel("Number of high-resolution shells",
                                  self, objectName="indent")
        self.number = QSpinBox()
        self.number.setMinimum(1)
        self.number.setValue(1)

        explicitToolTip = \
            "values in angstrom must be divided using commas without any " \
            "spaces and written in decreasing order"
        self.explicitRadio = QRadioButton(
            "Explicit definition of high-resolution shells")
        resolutionRadio.addButton(self.explicitRadio)
        self.explicitRadio.toggled.connect(self.resToggled)
        self.explicit = QLineEdit()
        self.explicit.setToolTip(explicitToolTip)
        self.explicit.setEnabled(False)
        self.explicitLabel = QLabel("e.g. 2.1,2.0,1.9",
                                    self, objectName="italic")
        self.explicitLabel.setToolTip("in angstrom")
        self.explicitLabel.setEnabled(False)

        self.res_init = QLineEdit()
        self.res_init.setToolTip("in angstrom")
        self.res_init.setValidator(doubleValidatorPositive)
        self.res_initLabel = QLabel("Starting resolution", self)
        self.res_initLabel.setToolTip("in angstrom")

        self.refinementLabel = QLabel("Refinement software", self)
        refinementRadio = QButtonGroup(self)
        self.refmacRadio = QRadioButton("REFMAC5", objectName="italic")
        refinementRadio.addButton(self.refmacRadio)
        self.refmacRadio.setChecked(True)
        self.refmacRadio.toggled.connect(self.refinementToggled)
        self.phenixRadio = QRadioButton("phenix.refine", objectName="italic")
        refinementRadio.addButton(self.phenixRadio)
        self.phenixRadio.setChecked(False)
        self.phenixRadio.toggled.connect(self.refinementToggled)

        self.projectName = QLineEdit("project")
        self.projectName.setValidator(alphanumValidator)
        self.projectLabel = QLabel("Project name", self)

        self.workdirName = QLineEdit(os.getcwd())
        self.workdirName.textChanged.connect(self.setworkdir)
        self.workdirName.setStyleSheet("background-color: lightgreen;")
        self.workdirButton = QPushButton("Working directory")
        self.workdirButton.clicked.connect(self.getworkdir)
        
        files = {}
        files["xyzin"] = {"label": "Structure model",
                          "filetype": "*.pdb *.mmcif *.cif"}
        files["hklin"] = {"label": "Diffraction data (merged)",
                          "filetype": "*.mtz"}
        files["hklin_unmerged"] = {"label": "Diffraction data (unmerged)",
                                   "filetype": "*.*"}
        files["libin"] = {"label": "Geometric restraints",
                          "filetype": "*.cif *.geo"}
        files["comin"] = {"label": "Command file for REFMAC5",
                          "filetype": "*.*"}
        files["def"] = {
            "label": "Keyword file for phenix.refine",
            "filetype": "*.def"}
        files["tlsin"] = {"label": "Input fixed TLS parameters",
                          "filetype": "*.tls *.tlsin *.tlsout"}
        
        self.xyzinFileName = QLineEdit(objectName="required")
        self.xyzinFileName.textChanged.connect(self.checkfile)
        self.xyzinButton = QPushButton(
            files["xyzin"]["label"], objectName="required")
        self.xyzinButton.clicked.connect(
            lambda: self.getfile(self.xyzinFileName, attrs=files["xyzin"]))

        self.hklinFileName = QLineEdit(objectName="required")
        self.hklinFileName.textChanged.connect(self.checkfile)
        self.hklinButton = QPushButton(
            files["hklin"]["label"], objectName="required")
        self.hklinButton.clicked.connect(
            lambda: self.getfile(self.hklinFileName, attrs=files["hklin"]))

        self.hklin_unmergedFileName = QLineEdit()
        self.hklin_unmergedFileName.textChanged.connect(self.checkfile)
        self.hklin_unmergedButton = QPushButton(
            files["hklin_unmerged"]["label"])
        self.hklin_unmergedButton.clicked.connect(
            lambda: self.getfile(self.hklin_unmergedFileName, attrs=files["hklin_unmerged"]))

        self.libinFileName = QLineEdit()
        self.libinFileName.textChanged.connect(self.checkfile)
        self.libinButton = QPushButton(
            files["libin"]["label"])
        self.libinButton.clicked.connect(
            lambda: self.getfile(self.libinFileName, attrs=files["libin"]))

        self.cominFileName = QLineEdit()
        self.cominFileName.textChanged.connect(self.checkfile)
        self.cominButton = QPushButton(
            files["comin"]["label"])
        self.cominButton.clicked.connect(
            lambda: self.getfile(self.cominFileName, attrs=files["comin"]))

        self.defFileName = QLineEdit()
        self.defFileName.textChanged.connect(self.checkfile)
        self.defButton = QPushButton(
            files["def"]["label"])
        self.defButton.clicked.connect(
            lambda: self.getfile(self.defFileName, attrs=files["def"]))
        self.defFileName.hide()
        self.defButton.hide()

        self.ncyc = QSpinBox()
        self.ncyc.setMinimum(1)
        self.ncyc.setValue(10)
        self.ncycLabel = QLabel("Number of refinement cycles",
                                self)
        self.ncyc.setToolTip("Number of refinement cycles that will "
                             "be performed in every resolution step")
        self.ncycLabel.setToolTip("Number of refinement cycles that will "
                                  "be performed in every resolution step")

        self.weightCheckBox = QCheckBox("Use automatic weighting")
        self.weightCheckBox.setChecked(True)
        self.weightCheckBox.stateChanged.connect(self.weight_show_hide)

        self.weightLabel = QLabel("Use weighting term")
        self.weightLabel.hide()
        self.weight = QLineEdit(objectName="required")
        self.weight.setValidator(doubleValidatorPositive)
        self.weight.textChanged.connect(self.checkweight)
        self.weight.hide()

        # TLS
        self.tlsCheckBox = QCheckBox("Input fixed TLS parameters")
        self.tlsCheckBox.stateChanged.connect(self.tls_show_hide)

        self.tlsinFileName = QLineEdit()
        self.tlsinFileName.textChanged.connect(self.checkfile)
        self.tlsinButton = QPushButton(
            files["tlsin"]["label"], objectName="indent")
        self.tlsinButton.clicked.connect(
            lambda: self.getfile(self.tlsinFileName, attrs=files["tlsin"]))

        self.tls_ncyc = QSpinBox()
        self.tls_ncyc.setMinimum(1)
        self.tls_ncyc.setValue(10)
        self.tls_ncycLabel = QLabel("Number of cycles of TLS refinement", self,
                                    objectName="indent")

        self.free = QLineEdit("0")
        self.free.setValidator(intValidatorPositive)
        self.freeLabel = QLabel("Exclude reflections with free-flag", self)

        self.tlsin_keep = QCheckBox("Keep using the same TLS input file "
                                    "in all the refinement runs",
                                    objectName="indent")

        tls_widgets = [self.tlsinFileName, self.tlsinButton, self.tls_ncyc,
                       self.tls_ncycLabel, self.tlsin_keep]
        for tls_widget in tls_widgets:
            tls_widget.hide()

        # Complete cross-validation, prerefinement, input model modification
        self.completeCheckBox = QCheckBox(
            "Perform complete cross-validation accross all the free-flag sets")
        self.completeCheckBox.stateChanged.connect(
            self.complete_show_hide)

        self.prerefinementCheckBox = QCheckBox(
            "Refine the input structure model at the starting "
            "resolution")
        self.prerefinementCheckBox.stateChanged.connect(
            self.prerefinement_show_hide)

        prerefinement_ncycToolTip = \
            "Number of refinement cycles to be performed as pre-" + \
            "refinement of the input structure model before paired " + \
            "refinement (the initial high resolution limit is " + \
            "used). Pre-refinement is performed by default in case " + \
            "of the complete cross-validation protocol."
        self.prerefinement_ncyc = QSpinBox()
        self.prerefinement_ncyc.setMinimum(1)
        self.prerefinement_ncyc.setValue(20)
        self.prerefinement_ncycLabel = QLabel(
            "Number of pre-refinement cycles", self, objectName="indent")
        self.prerefinement_ncyc.setToolTip(prerefinement_ncycToolTip)
        self.prerefinement_ncycLabel.setToolTip(prerefinement_ncycToolTip)

        prerefinement_shakeToolTip = \
            "Randomize coordinates of the input structure model " + \
            "with the given mean error value. This is done by " + \
            "default in the case of the complete cross-validation protocol."
        self.prerefinement_shakeCheckBox = QCheckBox(
            "Randomize coordinates with mean error", objectName="indent")
        self.prerefinement_shakeCheckBox.setToolTip(prerefinement_shakeToolTip)
        self.prerefinement_shake = QLineEdit("0.25")
        self.prerefinement_shake.setValidator(doubleValidatorPositive)
        self.prerefinement_shake.setToolTip(prerefinement_shakeToolTip)

        self.prerefinement_resetCheckBox = QCheckBox(
            "Reset B-factors to mean value", objectName="indent")
        self.prerefinement_resetCheckBox.stateChanged.connect(
            self.prerefinement_setUncheck)
        self.prerefinement_resetCheckBox.setToolTip(
            "Reset atomic B-factors of the input structure model to "
            "the mean value. This is done by default in the case of "
            "the completecross-validation protocol.")

        prerefinement_setToolTip = \
            "Set atomic B-factors of the input structure model to "
        "the given value (angstrom)"
        self.prerefinement_setCheckBox = QCheckBox(
            "Set atomic B-factors", objectName="indent")
        self.prerefinement_setCheckBox.setToolTip(prerefinement_setToolTip)
        self.prerefinement_setCheckBox.stateChanged.connect(
            self.prerefinement_resetUncheck)
        self.prerefinement_set = QLineEdit("30.0")
        self.prerefinement_set.setValidator(doubleValidatorPositive)
        self.prerefinement_set.setToolTip(prerefinement_setToolTip)

        prerefinement_addToolTip = \
            "Add the given value to B-factors of the input " + \
            "structure model (in angstrom)"
        self.prerefinement_addCheckBox = QCheckBox(
            "Add value to B-factors", objectName="indent")
        self.prerefinement_addCheckBox.setToolTip(prerefinement_addToolTip)
        self.prerefinement_addCheckBox.stateChanged.connect(
            self.prerefinement_setUncheck)
        self.prerefinement_add = QLineEdit("0")
        self.prerefinement_add.setValidator(doubleValidator)
        self.prerefinement_add.setToolTip(prerefinement_addToolTip)

        prerefinement_widgets = [
            self.prerefinement_ncyc, self.prerefinement_ncycLabel, 
            self.prerefinement_addCheckBox, self.prerefinement_add,
            self.prerefinement_setCheckBox, self.prerefinement_set,
            self.prerefinement_resetCheckBox,
            self.prerefinement_shakeCheckBox, self.prerefinement_shake]
        for prerefinement_widget in prerefinement_widgets:
            prerefinement_widget.hide()


        self.runButton = QPushButton("RUN")
        self.runButton.clicked.connect(self.run)
        self.stopButton = QPushButton("STOP")
        self.stopButton.clicked.connect(self.stop)
        self.stopButton.hide()

        self.browserLabel = QLabel("Job is running, open HTML log file ")
        self.browserLabel.hide()
        self.browserButton = QPushButton("Open in browser...")
        self.browserButton.hide()
        self.browserButton.clicked.connect(self.openBrowser)
        # myButton.move(150, 80)
        
        # self.setGeometry(96, 56, 406, 200)
        self.resize(800, 240)
        box_head = QVBoxLayout()
        grid_res = QGridLayout()
        grid_files = QGridLayout()
        grid_more = QGridLayout()
        grid_more2 = QGridLayout()
        grid_foot = QGridLayout()
        # grid.setSpacing(4)
        if os.path.isfile(logofilename):
            box_head.addWidget(self.logoLabel)
        box_head.addWidget(self.header)
        box_head.addStretch()
        grid_res.addWidget(self.step_numberRadio, 150, 0)
        grid_res.addWidget(self.stepLabel, 152, 0)
        grid_res.addWidget(self.step, 152, 2)
        grid_res.addWidget(self.numberLabel, 152, 4)
        grid_res.addWidget(self.number, 152, 6)
        grid_res.addWidget(self.explicitRadio, 156, 0)
        grid_res.addWidget(self.explicit, 156, 2)
        grid_res.addWidget(self.explicitLabel, 156, 4)
        grid_res.addWidget(self.res_initLabel, 180, 0)
        grid_res.addWidget(self.res_init, 180, 2)
        grid_res.addWidget(self.refinementLabel, 190, 0)
        grid_res.addWidget(self.refmacRadio, 190, 2)
        grid_res.addWidget(self.phenixRadio, 190, 4)
        grid_files.addWidget(self.projectLabel, 200, 0)
        grid_files.addWidget(self.projectName, 200, 2)
        grid_files.addWidget(self.workdirButton, 250, 0)
        grid_files.addWidget(self.workdirName, 250, 2)
        grid_files.addWidget(self.xyzinButton, 300, 0)
        grid_files.addWidget(self.xyzinFileName, 300, 2)
        grid_files.addWidget(self.hklinButton, 400, 0)
        grid_files.addWidget(self.hklinFileName, 400, 2)
        grid_files.addWidget(self.hklin_unmergedButton, 500, 0)
        grid_files.addWidget(self.hklin_unmergedFileName, 500, 2)
        grid_files.addWidget(self.libinButton, 600, 0)
        grid_files.addWidget(self.libinFileName, 600, 2)
        grid_files.addWidget(self.cominButton, 700, 0)
        grid_files.addWidget(self.cominFileName, 700, 2)
        grid_files.addWidget(self.defButton, 700, 0)
        grid_files.addWidget(self.defFileName, 700, 2)

        grid_more.addWidget(self.ncycLabel, 730, 0)
        grid_more.addWidget(self.ncyc, 730, 2)
        grid_more.addWidget(self.freeLabel, 730, 4)
        grid_more.addWidget(self.free, 730, 6)

        grid_more2.addWidget(self.weightCheckBox, 736, 0)
        grid_more2.addWidget(self.weightLabel, 740, 0)
        grid_more2.addWidget(self.weight, 740, 2)

        grid_more2.addWidget(self.tlsCheckBox, 750, 0)
        grid_more2.addWidget(self.tlsinButton, 760, 0)
        grid_more2.addWidget(self.tlsinFileName, 760, 2)
        grid_more2.addWidget(self.tls_ncycLabel, 761, 0)
        grid_more2.addWidget(self.tls_ncyc, 761, 2)
        grid_more2.addWidget(self.tlsin_keep, 762, 0, 1, 3)

        grid_more2.addWidget(self.completeCheckBox, 950, 0, 1, 3)
        grid_more2.addWidget(self.prerefinementCheckBox, 955, 0, 1, 3)
        grid_more2.addWidget(self.prerefinement_ncycLabel, 960, 0)
        grid_more2.addWidget(self.prerefinement_ncyc, 960, 2)
        grid_more2.addWidget(self.prerefinement_shakeCheckBox, 962, 0)
        grid_more2.addWidget(self.prerefinement_shake, 962, 2)
        grid_more2.addWidget(self.prerefinement_resetCheckBox, 964, 0)
        grid_more2.addWidget(self.prerefinement_setCheckBox, 965, 0)
        grid_more2.addWidget(self.prerefinement_set, 965, 2)
        grid_more2.addWidget(self.prerefinement_addCheckBox, 968, 0)
        grid_more2.addWidget(self.prerefinement_add, 968, 2)

        grid_foot.addWidget(self.runButton, 1000, 0)
        grid_foot.addWidget(self.stopButton, 1000, 1)
        grid_foot.addWidget(self.browserLabel, 1100, 0)
        grid_foot.addWidget(self.browserButton, 1100, 1)

        layout = box_head
        layout.addLayout(grid_res)
        layout.addLayout(grid_files)
        layout.addLayout(grid_more)
        layout.addLayout(grid_more2)
        layout.addLayout(grid_foot)
        self.setLayout(layout)
        self.show()

    def getworkdir(self):
        workdirname = QFileDialog.getExistingDirectory(self, 'Select directory')
        self.workdirName.setText(workdirname)

    def setworkdir(self):
        try:
            if os.path.isdir(self.workdirName.text()):
                os.chdir(str(self.workdirName.text()))
                self.workdirName.setStyleSheet("background-color: lightgreen;")
            else:
                self.workdirName.setStyleSheet("background-color: red;")
        except UnicodeDecodeError:
            self.workdirName.setStyleSheet("background-color: red;")
            
    def getfile(self, QLine, attrs):
        action = "Select " + attrs["label"].lower()
        filetype = attrs["label"] + " (" + attrs["filetype"] + ")" + \
            " ;; All files (*.*)"
        if "win" in platform.system().lower() or \
                "mac" in platform.system().lower():
            # file type filter does not work well
            filename = QFileDialog.getOpenFileName(
                self, action, "")
        else:
            filename = QFileDialog.getOpenFileName(
                self, action, "", filetype)
        QLine.setText(filename)

    def checkfile(self):
        try:
            if os.path.isfile(self.sender().text()):
                self.sender().setStyleSheet("background-color: lightgreen;")
            else:
                self.sender().setStyleSheet("background-color: red;")
        except UnicodeDecodeError:
            self.sender().setStyleSheet("background-color: red;")

    def checkweight(self):
        weight = str(self.weight.text())
        try:
            if float(weight) > 0:
                self.sender().setStyleSheet("background-color: lightgreen;")
            else:
                self.sender().setStyleSheet("background-color: red;")
        except:
            self.sender().setStyleSheet("background-color: red;")

    def refinementToggled(self):
        widgets_refmac = [
            self.cominButton, self.cominFileName, self.weightCheckBox,
            self.tlsCheckBox]
        widgets_refmac_weight = [self.weightLabel, self.weight]
        widgets_refmac_tls = [
            self.tlsinButton, self.tlsinFileName, self.tls_ncycLabel,
            self.tlsin_keep, self.tls_ncyc]
        widgets_phenix = [self.defButton, self.defFileName]
        if self.refmacRadio.isChecked():
            self.ncyc.setValue(10)
            if not self.weightCheckBox.isChecked():
                widgets_refmac += widgets_refmac_weight
            if self.tlsCheckBox.isChecked():
                widgets_refmac += widgets_refmac_tls
            for widget in widgets_refmac:
                widget.show()
            for widget in widgets_phenix:
                widget.hide()
        if self.phenixRadio.isChecked():
            self.ncyc.setValue(3)
            widgets_refmac += widgets_refmac_weight + widgets_refmac_tls
            for widget in widgets_refmac:
                widget.hide()
            for widget in widgets_phenix:
                widget.show()
            

    def resToggled(self):
        if self.step_numberRadio.isChecked():
            self.step.setEnabled(True)
            self.number.setEnabled(True)
            self.explicit.setEnabled(False)
            self.explicitLabel.setEnabled(False)
        else:  # self.explicitRadio.isChecked():
            self.explicit.setEnabled(True)
            self.step.setEnabled(False)
            self.number.setEnabled(False)
            self.explicitLabel.setEnabled(True)

    def weight_show_hide(self):
        if not self.weightCheckBox.isChecked():
            self.weightLabel.show()
            self.weight.show()
        else:  # self.weightCheckBox.isChecked()
            self.weightLabel.hide()
            self.weight.hide()

    def tls_show_hide(self):
        tls_widgets = [self.tlsinFileName, self.tlsinButton, self.tls_ncyc,
                       self.tls_ncycLabel, self.tlsin_keep]
        if self.tlsCheckBox.isChecked():
            for tls_widget in tls_widgets:
                tls_widget.show()
        else:
            for tls_widget in tls_widgets:
                tls_widget.hide()

    def complete_show_hide(self):
        if self.completeCheckBox.isChecked():
            self.prerefinementCheckBox.setChecked(True)
            self.prerefinement_show_hide  # show
            self.prerefinementCheckBox.setEnabled(False)
            self.free.setEnabled(False)
            self.freeLabel.setEnabled(False)
        else: 
            # not hide self.prerefinement_show_hide
            self.prerefinementCheckBox.setEnabled(True)
            self.free.setEnabled(True)
            self.freeLabel.setEnabled(True)
                
    def prerefinement_show_hide(self):
        prerefinement_widgets = [
            self.prerefinement_ncyc, self.prerefinement_ncycLabel, 
            self.prerefinement_addCheckBox, self.prerefinement_add,
            self.prerefinement_setCheckBox, self.prerefinement_set,
            self.prerefinement_resetCheckBox,
            self.prerefinement_shakeCheckBox, self.prerefinement_shake]
        if self.prerefinementCheckBox.isChecked():
            for prerefinement_widget in prerefinement_widgets:
                prerefinement_widget.show()
            self.prerefinement_resetCheckBox.setChecked(True)
            self.prerefinement_shakeCheckBox.setChecked(True)
        else:  # not self.prerefinementCheckBox.isChecked():
            for prerefinement_widget in prerefinement_widgets:
                prerefinement_widget.hide()

    def prerefinement_setUncheck(self):
        if self.prerefinement_resetCheckBox.isChecked() or \
                self.prerefinement_addCheckBox.isChecked():
            self.prerefinement_setCheckBox.setChecked(False)

    def prerefinement_resetUncheck(self):
        if self.prerefinement_setCheckBox.isChecked():
            self.prerefinement_resetCheckBox.setChecked(False)
            self.prerefinement_addCheckBox.setChecked(False)

    def run(self):
        # Check validity of required arguments and working directory
        if not os.path.isfile(str(self.xyzinFileName.text())) or \
                not os.path.isfile(str(self.hklinFileName.text())) or \
                not os.path.isdir(str(self.workdirName.text())):
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setText("Structure model, merged diffracion data and "
                        "working directory must be valid!")
            # msg.setInformativeText("This is additional information")
            msg.setWindowTitle("PAIREF Error")
            msg.exec_()
            return
        # Prepare arguments
        argms = ["--XYZIN", str(self.xyzinFileName.text())]
        argms += ["--HKLIN", str(self.hklinFileName.text())]
        if self.hklin_unmergedFileName.text():
            argms += ["-u", str(self.hklin_unmergedFileName.text())]
        if self.libinFileName.text():
            argms += ["--LIBIN", str(self.libinFileName.text())]
        if self.cominFileName.text() and self.refmacRadio.isChecked():
            argms += ["--comfile", str(self.cominFileName.text())]
        elif self.defFileName.text() and self.phenixRadio.isChecked():
            argms += ["--def", str(self.defFileName.text())]
        if self.tlsinFileName.text() and self.refmacRadio.isChecked():
            argms += ["--TLSIN", str(self.tlsinFileName.text())]
        if self.refmacRadio.isChecked():
            argms += ["--refmac"]
        elif self.phenixRadio.isChecked():
            argms += ["--phenix"]
        if self.res_init.text() :
            argms += ["-i", str(self.res_init.text())]
        if self.step_numberRadio.isChecked():
            if self.step.text() :
                argms += ["-s", str(self.step.text())]
            if self.number.text():
                argms += ["-n", str(self.number.text())]
        else:  # self.explicitRadio is Checked:
            argms += ["-r", str(self.explicit.text())]
        if self.projectName.text():
            argms += ["-p", str(self.projectName.text())]
        if self.ncyc.value():
            argms += ["--ncyc", str(self.ncyc.value())]
        if self.free.text() and not self.completeCheckBox.isChecked():
            argms += ["--flag", str(self.free.text())]
        if not self.weightCheckBox.isChecked() and self.refmacRadio.isChecked():
            # e.i. not auto weight
            argms += ["--weight", str(self.weight.text())]
        if self.tls_ncyc.value() and self.tlsCheckBox.isChecked() and \
                self.refmacRadio.isChecked():
            argms += ["--TLS-ncyc", str(self.tls_ncyc.value())]
        if self.tlsin_keep.isChecked() and self.tlsCheckBox.isChecked() and \
                self.refmacRadio.isChecked():
            argms += ["--TLSIN-keep"]
        if self.completeCheckBox.isChecked():
            argms += ["--complete"]
        if self.prerefinementCheckBox.isChecked():
            argms += ["--prerefinement-ncyc", str(self.prerefinement_ncyc.text())]
        if self.prerefinement_shakeCheckBox.isChecked() and self.prerefinementCheckBox.isChecked():
            argms += ["--prerefinement-shake-sites", str(self.prerefinement_shake.text())]
        if self.prerefinement_resetCheckBox.isChecked() and self.prerefinementCheckBox.isChecked():
            argms += ["--prerefinement-reset-bfactor"]
        if self.prerefinement_setCheckBox.isChecked() and self.prerefinementCheckBox.isChecked():
            argms += ["--prerefinement-set-bfactor", str(self.prerefinement_set.text())]
        if self.prerefinement_addCheckBox.isChecked() and self.prerefinementCheckBox.isChecked():
            argms += ["--prerefinement-add-to-bfactor", str(self.prerefinement_add.text())]
        if self.prerefinementCheckBox.isChecked() and \
                not self.prerefinement_shakeCheckBox.isChecked() and \
                not self.prerefinement_resetCheckBox.isChecked() and \
                not self.prerefinement_setCheckBox.isChecked() and \
                not self.prerefinement_addCheckBox.isChecked():
            argms += ["--prerefinement-no-modification"]
        # Predict new working directory
        project = str(self.projectName.text())
        if project == "":
            project = "project"
        workdir = "pairef_" + project
        while os.path.isdir(workdir):
            workdir += "_new"
        self.logfilepath = os.path.abspath(workdir) + \
            os.path.sep + "PAIREF_" + project + ".html"
        # pairef.run_pairef(argms)
        # import subprocess
        # subprocess.call([sys.executable, "-m", "pairef"] + argms)
        # Run
        argms = ["-m", "pairef"] + argms
        if "win" in platform.system().lower() or "mac" in platform.system().lower():
            ccp4_command = sys.executable
        else:
            from .preparation import which
            if which("ccp4-python"):
                ccp4_command = "ccp4-python"
            else:
                ccp4_command = sys.executable
        process = QProcess()
        print("pairef-gui: executing command:\n", ccp4_command, argms)
        self.re, self.pid = process.startDetached(ccp4_command, argms, ".")
        # except SystemExit:
        #     print("ignoring SystemExit")
        # Show/hide/enable buttons
        self.browserLabel.setText(
            "Job was executed, check current status in log file\n" + self.logfilepath)
        self.stopButton.show()
        self.stopButton.setEnabled(True)
        self.browserLabel.show()
        self.browserButton.show()

    def openBrowser(self):
        webbrowser.open("file://" + self.logfilepath)

    def stop(self):
        os.kill(self.pid, signal.SIGTERM)
        self.sender().setEnabled(False)
        message = self.browserLabel.text()
        message.replace("executed", "terminated")
        self.browserLabel.setText(message)


def gui():
    """Opens the window of PAIREF graphical interface (in PyQt4)"""
    app = QApplication(sys.argv)
    w = MyWindow()
    app.exec_()



# coding: utf-8
from __future__ import print_function
import sys
import os
import subprocess
import datetime
from math import sqrt, pow
from collections import OrderedDict  # Python 2.7
from .settings import warning_dict, date_time, settings
from .commons import twodec, twodecname, fourdec, extract_from_file
from .commons import warning_my


BINS_LOW = 10


def which(program):
    """Checks if `program` exists and finds its location. Analogy of the
    `which` GNU/Linux command.

    Args:
        program (str): Name of an executable

    Returns:
        str: Path of an executable location
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program) or is_exe(program + ".exe"):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
            exe_file = os.path.join(path, program + ".exe")
            if is_exe(exe_file):
                return exe_file
    return None


def welcome(args, pairef_version):
    '''Print introduction information about the module and
    input parameters.

    Args:
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()
        pairef_version (str)

    Returns:
        bool: True
    '''
    import getpass
    import socket


    print("""
 __   _  ___ __  ___ ___
 )_) /_)  )  )_) )_  )_
/   / / _(_ / \ (__ (
""")
    print("automatic PAIRed REFinement protocol")
    print("version: " + pairef_version)
    print("run date and time: " + date_time)
    print("user@host: " + getpass.getuser() + "@" + socket.gethostname())
    print("")
    print('Please cite: "Paired refinement under the control of PAIREF"')
    print("M. Maly, K. Diederichs, J. Dohnalek, P. Kolenko (2020) IUCrJ 7")
    print("")
    print("Command line arguments: " + " ".join(sys.argv[1:]))
    print("")
    print("Program has been executed with following input parameters:")
    if args.refmac:
        print(" * Refinement software: REFMAC5")
    if args.phenix:
        print(" * Refinement software: phenix.refine")
    print(" * XYZIN: " + args.xyzin)
    print(" * HKLIN: " + args.hklin)
    if args.hklin_unmerged:
        print(" * HKLIN unmerged: " + args.hklin_unmerged)
    if args.libin:
        print(" * LIBIN: " + args.libin)
    if args.tlsin:
        print(" * TLSIN: " + args.tlsin)
    if args.tlsin_keep:
        print("   (keep using the same TLS input file in all the refinement "
              "runs)")
    print(" * Project name: " + str(args.project))
    # if args.step:
    #     print(" * Resolution step in angstroem: " + str(args.step))
    # if args.n_shells:
    #     print(" * Number of res. shells: " + str(args.n_shells))
    if args.res_shells:
        print(" * Resolution shells: " + str(args.res_shells))
    if args.weight:
        print(" * Weight matrix: " + str(args.weight))
    if args.tls_ncyc:
        print(" * Number of number of cycles of TLS refinement: "
              "" + str(args.tls_ncyc))
    if args.ncyc:
        print(" * Number of refinement cycles that will be performed in "
              "every resolution step: " + str(args.ncyc))
    if args.prerefinement_ncyc:
        print(" * Number of pre-refinement cycles that will be performed "
              "before the paired refinement protocol: "
              "" + str(args.prerefinement_ncyc))
    if args.complete_cross_validation:
        print(" * Complete cross-validation will be performed.")

    if (args.complete_cross_validation or args.no_modification or
            args.reset_bfactor or args.add_to_bfactor or args.set_bfactor or
            args.shake_sites):
        print(" * Modification of the input structure model:")
    if args.reset_bfactor:
        print("   - Reset B-factors to the mean value")
    if args.add_to_bfactor:
        print("   - Add value to B-factors: " +
              twodec(args.add_to_bfactor))
    if args.set_bfactor:
        print("   - Set B-factors to the value: " + twodec(args.set_bfactor))
    if args.shake_sites:
        print("   - Randomize coordinates with the "
              "given mean error value: " + twodec(args.shake_sites))

    if args.constant_grid:
        print(" * The same FFT grid will be kept through the whole paired "
              "refinement.")
    if args.comin:
        print(" * Com file for REFMAC5: " + args.comin)
    if args.defin:
        print(" * Keyword file for phenix.refine: " + args.defin)
    if args.test:
        print(" * Light-testing mode (REFMAC5 will not be executed).")
    print("")
    return True


def create_workdir(project):
    '''Create directory `pairef_`project``, if exists, create directory
    `pairef_`project`_new` (recursively).

    Args:
        project (str): Name of the project

    Returns:
        str: Name of the created working directory
    '''
    workdir = "pairef_" + project
    # Make directory for the project
    try:
        os.mkdir(workdir)
    except OSError:
        warning_my("workdir",
                   "This directory contains directory "
                   "`pairef_" + project + "` "
                   "already, choosing another directory name.")
        workdir = create_workdir(project + "_new")
    return workdir


class output_log:
    "Set to write `sys.stdout` to screen and also in file `PAIREF_out.log`."
    # Not working for STDERR! TODO! write on internet...
    def __init__(self, stdout, filename):
        self.stdout = stdout
        self.logfile = open(filename, 'a')

    def write(self, text):
        self.stdout.write(text)
        self.logfile.write(text)

    def close(self):
        self.stdout.close()
        self.logfile.close()

    def flush(self):
        self.stdout.close()
        self.logfile.close()


def def_res_shells(args, refinement, res_high_mtz, res_low=999):
    """Determine high resolution shells and number of low resolution bins.

    The first value of list should be the resolution of data which
    were used for the refinement of the input structure model.
    If explicit definition (args.res_shells) is set, test its correctness.
    If it is valid, use, if not, define it automatically (shell step 0.05 A)

    Args:
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()
        refinement (str): "refmac" or "phenix"
        res_high_mtz (float): High resolution diffraction limit of args.hklin
        res_low (float): Low resolution diffraction limit of args.hklin

    Returns:
        (tuple):
            * shells (*list*)
            * n_bins_low (*int*)
            * n_flag_sets (*int*)
            * default_shells_definition (*bool*)
    """
    # If the resolution of input model == resolution of diffr. data, abort
    if twodec(args.res_init) == twodec(res_high_mtz):
        sys.stderr.write("ERROR: The given input structure model "
                         "" + args.xyzin + " was refined at the same "
                         "resolution as the given diffraction data "
                         "" + args.hklin + " have. Nothing to do."
                         "\nAborting.\n")
        sys.exit(1)
    # Estimation of
    #   1. a number of low resolution bins and
    #   2. a number of free reflection sets
    # If not successful, just divide into 12 resolution bins
    # and consider 20 free reflection sets.
    n_i_obs = 0
    n_i_obs_low = 0
    n_flag_sets = 0
    if refinement == "refmac":
        # using `mtzdump` from CCP4 to get both n_i_obs_low and n_flag_sets
        tool = "mtzdump"
        p = subprocess.Popen(["mtzdump", "HKLIN", args.hklin],
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             shell=settings["sh"])
        com = "STATS NBIN 1 RESO " + twodec(res_low) + " " \
            "" + twodec(args.res_init) + "\n end\n"
        output, err = p.communicate(com)
        # rc = p.returncode
        try:
            for line in output.splitlines():
                if " * Number of Reflections = " in line:
                    n_i_obs = int(line.split()[-1])
                elif len(line.split()) == 12:
                    if "free" in line.split()[-1].lower():
                        n_flag_sets = int(float(line.split()[3])) - \
                            int(float(line.split()[2])) + 1
                elif " No. of reflections used in FILE STATISTICS" in line:
                    n_i_obs_low = int(line.split()[-1])
                    break
        except ValueError:
            n_bins_low = 12
            warning_my("mtzdump",
                       "Definition of binning "
                       "and search for number of sets of free reflection "
                       "using mtzdump were not successful. "
                       "Data will be divided into " + str(n_bins_low) + " "
                       "resolution bins. "
                       "Is the input MTZ file " + args.hklin + " OK?")
    elif refinement == "phenix":
        from iotbx.reflection_file_reader import any_reflection_file
        from iotbx import mtz
        import cStringIO
        tool = "CCTBX"
        # 1. get n_i_obs_low using CCTBX
        n_i_obs_low_list = []
        n_i_obs_list = []
        hkl_in = any_reflection_file(file_name=args.hklin)
        miller_arrays = hkl_in.as_miller_arrays()
        for column in miller_arrays:
            if "xray" in str(column.observation_type()):
                n_i_obs_list.append(column.size())
                column_res_init = column.resolution_filter(
                    d_max=float(res_low), d_min=float(args.res_init))
                n_i_obs_low_list.append(column_res_init.size())
        n_i_obs = min(n_i_obs_list)
        n_i_obs_low = min(n_i_obs_low_list)
        # 2. get n_flag_sets using CCTBX
        mtz_object = mtz.object(file_name=args.hklin)
        out = cStringIO.StringIO()
        mtz_object.show_summary(out=out)
        out = out.getvalue()
        out_lines = out.splitlines()
        for i in range(len(out_lines)):
            if "free" in out_lines[i].lower():
                j = i
        if "j" in vars():
            n_flag_sets = int(float(out_lines[j].split()[4])) - \
                int(float(out_lines[j].split()[3])) + 1

    if n_flag_sets == 0:
        n_flag_sets = 20
        if args.complete_cross_validation:
            warning_my("flags",
                       "Search for number of sets of free reflection "
                       "using " + tool + " was not successful. "
                       "It will be assumed that 20 sets of free reflection "
                       "are present in the input MTZ file " + args.hklin + ".")
    else:
        if args.complete_cross_validation:
            print(str(n_flag_sets) + " sets of free reflection were found and "
                  "will be used in complete cross-validation.")
    if n_i_obs == 0 or n_i_obs_low == 0:
        n_bins_low = 12
        warning_my("binning",
                   "Definition of binning using " + tool + " was not success"
                   "ful. Data will be divided into " + str(n_bins_low) + " "
                   "resolution bins. "
                   "Is the input MTZ file " + args.hklin + " OK?")
    else:  # If run of mtzdump was succesful
        n_bins_low_start = 12
        n_i_obs_thinner_shell_start = 0
        n_bins_low = n_bins_low_start
        n_i_obs_thinner_shell = n_i_obs_thinner_shell_start
        while n_i_obs_thinner_shell < 2000:
            n_bins_low = n_bins_low - 1
            n_i_obs_thinner_shell = \
                sqrt(2) * n_i_obs_low / pow(n_bins_low, 1.5)
    # Now `n_bins_low` is ready

    if args.res_shells:
        # Explicit definition of high resolution shells
        default_shells_definition = False
        shells_high = args.res_shells.split(",")
        for i in range(len(shells_high)):
            try:
                shells_high[i] = round(float(shells_high[i]), 2)
            except ValueError:
                sys.stderr.write(
                    "ERROR: Explicit definition of high "
                    "resolution shells (option -r) is not correct. ")
                sys.stderr.write(
                    "Values must be divided using commas without "
                    "any spaces (e.g. 2.1,2.0,1.9).")
                sys.stderr.write("\nAborting.\n")
                sys.exit(1)
            if i != 0:
                if float(shells_high[i - 1]) <= float(shells_high[i]):
                    sys.stderr.write(
                        "ERROR: Explicit definition of high "
                        "resolution shells (option -r) is not correct. ")
                    sys.stderr.write(
                        "Values must be set in the decreasing order "
                        "(e.g. 2.1,2.0,1.9).")
                    sys.stderr.write("\nAborting.\n")
                    sys.exit(1)
        if shells_high[0] >= args.res_init:
            sys.stderr.write(
                "ERROR: Explicit definition of high "
                "resolution shells (option -r) is not correct. ")
            sys.stderr.write(
                "Resolution of the first shell (" + twodec(shells_high[0]) + ""
                " A), "
                "which was explicitely defined, is lower than or the same as "
                "the resolution of data which were used for refinement of the "
                "input structure model (" + twodec(args.res_init) + " A).")
            sys.stderr.write("\nAborting.\n")
            sys.exit(1)

    elif args.step and args.n_shells:
        # Defined step (in A) and number of shells to be added
        default_shells_definition = False
        res_fin = args.res_init - args.step * args.n_shells
        if res_fin < 0:
            sys.stderr.write("ERROR: Current setup of resolution shells is "
                             "not valid. High resolution limit of the last "
                             "shell would be negative.\n")
            sys.stderr.write("\nRESOLUTION_INITIAL - (RESOLUTION_STEP * NUM"
                             "BER_OF_SHELLS) = " + twodec(args.res_init) + ""
                             " - ( " + str(args.step) + " * "
                             "" + str(args.n_shells) + ""
                             " ) = " + twodec(res_fin) + " < 0")
            sys.stderr.write("\nAborting.\n")
            sys.exit(1)
        shells_high = []
        for i in range(args.n_shells):
            shells_high.append(args.res_init - (i + 1) * args.step)

    else:
        # Default setting - 0.05A wide high resolution shells
        shell_width = 0.05
        default_shells_definition = True
        shells_high = []
        res_cur = args.res_init - shell_width
        while round(res_cur, 3) >= round(res_high_mtz, 3):
            shells_high = shells_high + [res_cur]
            res_cur = res_cur - shell_width
        shells_high = shells_high[:-1]
        shells_high = shells_high + [res_high_mtz]

        # Automatic estimation of high res. shell (using `mtzdump` from CCP4)
        # (shells with increasing number of reflections)
        # default_shells_definition = True
        # n_i_obs_high = n_i_obs - n_i_obs_low
        # n_i_obs_remaining = n_i_obs_high
        # n_bins_high = 0
        # i = 1
        # n_i_obs_shells_high = n_i_obs_thinner_shell * sqrt(n_bins_low + i)
        # while n_i_obs_remaining > n_i_obs_shells_high:
            # i += 1
            # n_i_obs_shell_high = n_i_obs_thinner_shell * sqrt(n_bins_low + i)
            # n_i_obs_remaining = n_i_obs_remaining - n_i_obs_shell_high
            # n_bins_high += 1
        # p = subprocess.Popen(["mtzdump", "HKLIN", args.hklin],
                             # stdin=subprocess.PIPE,
                             # stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # com = "STATS NBIN " + str(n_bins_high) + " RESO " \
            # "" + twodec(args.res_init) + "" \
            # " " + twodec(res_high_mtz) + "\n end\n"
        # output, err = p.communicate(com)
        # shells_high = []
        # try:
            # for line in output.splitlines():
                # if " PARTIAL FILE STATISTICS for resolution range" in line:
                    # shells_high.append(
                        # round(1/sqrt(float(line.split()[-1])), 2))
                    # #  there is no break
        # except:
            # sys.stderr.write(
                # "ERROR: Definition of high resolution shells  using mtzdump "
                # "was not successful. Is the MTZ file OK?")
            # sys.stderr.write(
                # "Use manual definition of the high resolution shells to "
                # "overcome this problem (option -r).")
            # sys.stderr.write("\nAborting.\n")
            # sys.exit(1)

    shells = [args.res_init] + shells_high
    return shells, n_bins_low, n_flag_sets, default_shells_definition


def res_high_from_xyzin(xyzin, format=".pdb"):
    """Finds a line containing `RESOLUTION RANGE HIGH` in the file `xyzin`,
    picks the last word of the line (that should be the high resolution)
    and rounds it to two decimals. If it is not successful, returns -1.

    Args:
        xyzin (str): Input structure model (PDB or mmCIF format)

    Returns:
        float: initial high resolution limit rounded to two decimals or -1 if \
               it was not found
    """
    if "cif" in format:
        res_high = extract_from_file(
            filename=xyzin, searched="_refine.ls_d_res_high", skip_lines=0,
            n_lines=1, nth_word=-1, not_found='N/A', get_first=True)
    else:  # pdb
        res_high = extract_from_file(
            filename=xyzin, searched="RESOLUTION RANGE HIGH", skip_lines=0,
            n_lines=1, nth_word=-1, not_found='N/A', get_first=True)
    if res_high == 'N/A':
        return -1
    try:
        res_high = round(float(res_high), 2)
    except ValueError:
        return -1
    return res_high


def res_from_mtz(hklin):
    """Finds the low resolution diffraction limit of data `hklin` using CCTBX.

    Args:
        hklin (str): Name of diffraction data MTZ file

    Returns:
        (tuple):
            * res_low (*float*): Low resolution diffraction limit
            * res_high (*float*): High resolution diffraction limit
    """
    from iotbx.reflection_file_reader import any_reflection_file
    hkl_in = any_reflection_file(file_name=hklin)
    miller_arrays = hkl_in.as_miller_arrays()
    res_low = None
    res_high = None
    for column in miller_arrays:
        if "xray" in str(column.observation_type()):
            res_low = column.d_max_min()[0]
            res_high = column.d_max_min()[1]
            break
    return res_low, res_high


def res_from_hklin_unmerged(hklin_unmerged):
    """
    Finds a resolution range for given unmerged diffr. data file 
    `hklin_unmerged`.
    In the case of an ASCII file from XDS, find it by
    searching the option `INCLUDE_RESOLUTION_RANGE` in the file.
    In the case of an MTZ file, find it using `xia2`.
    In the case of a SCA file, do not check.

    Args:
        hklin_unmerged (str): Name of the unmerged diffr. data file

    Returns:
        (tuple):
            * res_low (*float*): Low resolution diffraction limit
            * res_high (*float*): High resolution diffraction limit
    """
    res_high_from_hklin_unmerged = None
    res_low_from_hklin_unmerged = None
    # may be an MTZ file
    if  "mtz" in hklin_unmerged.split(".")[-1].lower():
        ### Solution 1 - CCTBX
        ### res_low_from_hklin_unmerged, res_high_from_hklin_unmerged = \
        ###     res_from_mtz(hklin_unmerged)  # This needs to much RAM...

        ## Solution 2 - mtzdump
        ## p = subprocess.Popen(["mtzdump", "HKLIN", hklin_unmerged],
        ##                      stdin=subprocess.PIPE,
        ##                      stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        ## com = "end\n"
        ## output, err = p.communicate(com)
        ## for i in range(len(output.splitlines())):
        ##     if " *  Resolution Range :" in output.splitlines()[i]:
        ##         res_low_from_hklin_unmerged = \
        ##             float(output.splitlines()[i + 2].split()[-5])
        ##         res_high_from_hklin_unmerged = \
        ##             float(output.splitlines()[i + 2].split()[-3])
        ##         break

        # Solution 3 - xia2 - works with both CCP4 and phenix
        from xia2.Wrappers.CCP4.Mtzdump import Mtzdump
        m = Mtzdump()
        m.set_hklin(hklin_unmerged)
        m.dump()
        res_low_from_hklin_unmerged, res_high_from_hklin_unmerged = \
            m.get_resolution_range()
        
    # may be an XDS ASCII file
    elif "hkl" in hklin_unmerged.split(".")[-1].lower():
        hklin_unmerged_head = []
        with open(hklin_unmerged, "r") as hklin_unmerged_file:
            hklin_unmerged_head = \
                [next(hklin_unmerged_file) for x in xrange(1000)]
        for line in hklin_unmerged_head:
            if "INCLUDE_RESOLUTION_RANGE" in line:
                res_high_from_hklin_unmerged = float(line.split()[-1])
                res_low_from_hklin_unmerged = float(line.split()[-2])
                break
    if not res_high_from_hklin_unmerged or not res_low_from_hklin_unmerged:
        warning_my("merging_stats",
                   "The check of a resolution range of the unmerged data "
                   "file " + hklin_unmerged + " was not successful. ")
        res_low_from_hklin_unmerged = float("inf")
        res_high_from_hklin_unmerged = 0
    return res_low_from_hklin_unmerged, res_high_from_hklin_unmerged


def check_refinement_software(args, versions_dict, refinement="refmac"):
    """
    Check if the input structure model `args.xyzin` was refined in REFMAC5 
    or phenix.refine and in which version (PDB and mmCIF format is accepted).
    If it was not refined or if it was refined in another
    version of REFMAC5 or phenix.refine than is now installed, write warning.

    Args:
        args.xyzin (str): Input structure model (PDB or mmCIF format)
        versions_dict (dict): Dictionary containing a key `refmac_version` \
                              or `phenix_version`.
        refmac_version_installed (str): Version of REFMAC5 or phenix.refine \
                                        that is installed

    Returns:
        str: version of REFMAC5 or phenix.refine  that was used for the \
             refinement of `args.xyzin`, if it was not found, "N/A" is returned
    """
    if refinement == "refmac":
        refinement_name = "REFMAC5"
        version_installed = versions_dict["refmac_version"]
        if "cif" in settings["pdbORmmcif"]:
            refmac_confirm = extract_from_file(
                filename=args.xyzin, searched="_software.name", skip_lines=0,
                n_lines=1, nth_word=-1, not_found="N/A", get_first=True)
            if refmac_confirm == "refmac":
                version_xyzin = extract_from_file(
                    filename=args.xyzin, searched="_software.version",
                    skip_lines=0, n_lines=1, nth_word=-1, not_found="N/A",
                    get_first=True)
                version_xyzin = version_xyzin.replace("'", "")
            else:
                version_xyzin = "N/A"
        else:  # pdb
            version_xyzin = extract_from_file(
                filename=args.xyzin, searched="REFMAC", skip_lines=0,
                n_lines=1, nth_word=-1, not_found="N/A", get_first=True)
    elif refinement == "phenix":
        refinement_name = "phenix.refine"
        version_installed = versions_dict["phenix_version"]
        if "cif" in settings["pdbORmmcif"]:
            phenix_confirm = extract_from_file(
                filename=args.xyzin, searched="phenix.refine", skip_lines=0,
                n_lines=1, nth_word=1, not_found="N/A", get_first=True)
            if phenix_confirm == "phenix.refine":  # TO DO check
                version_xyzin = extract_from_file(
                    filename=args.xyzin, searched="phenix.refine",
                    skip_lines=0, n_lines=1, nth_word=2, not_found="N/A",
                    get_first=True)
            else:
                version_xyzin = "N/A"          
        else:  # pdb
            version_xyzin = extract_from_file(
                filename=args.xyzin,
                searched="REMARK   3   PROGRAM     : PHENIX",
                skip_lines=0, n_lines=1, not_found="N/A", get_first=True)
            version_xyzin = version_xyzin[0]
            if version_xyzin != "N/A":
                version_xyzin = version_xyzin.split()[5]
                version_xyzin = version_xyzin.replace("(", "")
                version_xyzin = version_xyzin.replace(")", "")
        if version_xyzin != "N/A":
            version_xyzin = version_xyzin.split("_")[0]
    if args.prerefinement_ncyc:  # TO DO is it correct?
        return version_xyzin
    if version_xyzin == "N/A":
        warning_my("refinement_not_before",
                   "The input structure model `" + args.xyzin + "` seems not "
                   "to be refined in " + refinement_name + ". "
                   "The obtained results could be misleading. "
                   "Consider refinement of the structure model "
                   "" + args.xyzin + " in " + refinement_name + ". This could be "
                   "performed using the arguments --" + refinement + " "
                   "--prerefinement-ncyc")
    else:
        if version_xyzin != version_installed:
            warning_my("refinement_version_mismatch",
                       "The version of " + refinement_name + " used for "
                       "refinement of the  input structure model "
                       "`" + args.xyzin + "` ("
                       "" + version_xyzin + ") is not the same as the "
                       "version of " + refinement_name + " that is now "
                       "installed and used during paired refinement ("
                       "" + version_installed + "). "
                       "The results from PAIREF could be misleading. "
                       "Consider refinement of the structure model "
                       "`" + args.xyzin + "` using the installed version of"
                       " " + refinement_name + ". This could be performed "
                       "using the arguments --" + refinement + " "
                       "--prerefinement-ncyc")
    return version_xyzin


def res_opt(shell, args, refinement="refmac"):
    """
    Finds optical resolution running command
    :code:`sfcheck -f project_Rflag_shellA.mtz -m project_Rflag_shellA.pdb`
    and writes the gained value into project_Optical_resolution.csv
    PDB format is required, does not work with mmCIF.

    Args:
        shell (float): Current resolution shell
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()

    Returns
        float: optical resolution
    """
    prefix = args.project + "_R" + str(args.flag).zfill(2) + "_" + \
        twodecname(shell) + "A"
    if refinement == "phenix":
        prefix += "_001"
    hklin = prefix + ".mtz"
    xyzin = prefix + ".pdb"
    command = ["sfcheck", "-f", hklin, "-m", xyzin]
    logfilename = prefix + "_sfcheck.out"
    with open(logfilename, "w") as logfile:
        p = subprocess.Popen(
            command, stdout=logfile, stderr=logfile, shell=settings["sh"])
        out, err = p.communicate()

    res_opt = 0
    with open(logfilename, "r") as logfile:
        for line in logfile.readlines():
            if "Optical Resolution" in line:
                res_opt = float(line.split()[-1])
                break
    if os.path.isfile("sfcheck.xml"):
        os.rename("sfcheck.xml", prefix + "_sfcheck.xml")
    if os.path.isfile("sfcheck_xxxx.ps"):
        os.rename("sfcheck_xxxx.ps", prefix + "_sfcheck.ps")
    if os.path.isfile("sfcheck.log"):
        os.rename("sfcheck.log", prefix + "_sfcheck.log")
    # TODO: warning
    # TODO help
    csvfilename = args.project + "_Optical_resolution.csv"
    if not os.path.isfile(csvfilename):
        with open(csvfilename, "w") as csvfile:
            csvfile.write("# Nominal resolution          Optical resolution\n")
    with open(csvfilename, "a") as csvfile:
        csvfile.write(
            twodec(shell) + 26 * " " + twodec(res_opt) + "\n")
    return float(twodec(res_opt))


def calculate_merging_stats(hklin_unmerged, shells, project, bins_low,
                            res_low_from_hklin_unmerged=float("inf"),
                            res_high_from_hklin_unmerged=0):
    """
    For given file `hklin_unmerged`, calculate the merging statistics
    using CCTBX.

    Args:
        hklin_unmerged (str): Name of the unmerged diffraction data file
        shells (list)
        project (str): Name of the project
        bins_low (list)
        costant_reflections_in_bins (bool)

    Returns:
        str: Name of a CSV file where the calculated statistics has been saved
    """
    print("\n     * Calculating merging statistics...", end="")

    ## Calculate statistics
    def calculate_merging_stats_run_cctbx(project, hklin, res_high,
                                          res_low=None, n_bins=None):
        import iotbx.merging_statistics
        i_obs = iotbx.merging_statistics.select_data(file_name=hklin,
                                                     data_labels=None)
        result = iotbx.merging_statistics.dataset_statistics(
            i_obs=i_obs,
            # crystal_symmetry=symm,
            d_min=res_high,   # My change
            d_max=res_low,    # My change
            n_bins=n_bins,
            log=None
            )
        logfilename = project + "_merging_stats_" + twodecname(res_high) + "" \
            "A.log"
        with open(logfilename, "w") as logfile:
            result.show(out=logfile, header=False)
        return True

    bins_total_proposed = bins_low + shells
    bins_total = []
    for i in range(len(bins_total_proposed)-1):
        if res_low_from_hklin_unmerged < bins_total_proposed[i + 1] \
                or res_high_from_hklin_unmerged > bins_total_proposed[i]:
            warning_my("merging_stats",
                       "Merging statistics of the data in "
                       "bin " + twodec(bins_total_proposed[i]) + "-"
                       "" + twodec(bins_total_proposed[i + 1]) + " A "
                       "could not be calculated as the unmerged data "
                       "(file " + hklin_unmerged + ") do not contain "
                       "data in this resolution range.")
        else:
            bins_total.append(bins_total_proposed[i])
            bins_total.append(bins_total_proposed[i + 1])
            # Remove duplicates from bins_total and keep order
            bins_total = list(OrderedDict.fromkeys(bins_total))
            calculate_merging_stats_run_cctbx(
                project, hklin_unmerged, res_high=bins_total[i + 1],
                res_low=bins_total[i], n_bins=1)
        print(" .", end="")
    print("")

    # Prepare a csv file header
    csvfilename = project + "_merging_stats.csv"
    with open(csvfilename, "w") as csvfile:
        csvfile.writelines("#shell d_max  d_min   #obs  #uniq   mult.  %comp"
                           "       <I>  <I/sI>    r_mrg   r_meas    r_pim   "
                           "r_anom   cc1/2   cc_ano     cc* \n")

    ## Collect statistics, calculate CC*-values and save them to CSV file

    def calculate_CCstar(lines, shell=1):
        """For each string of lines array, find 11th word which should be
        value of CChalf and calculate CC*. Put the result into lines
        array. Add number of bin to lines array.

        Args:
            lines (list): list of strings
            shell (int)

        Returns:
            list
        """
        for i in range(len(lines)):
            # Compute CC*
            bin_CChalf = lines[i].split()[-2]
            if bin_CChalf < 0:
                bin_CCstar = "N/A"
                warning_my("CC*", "A CC*-value for a particular shell "
                           "could not be calculated as it is undefined "
                           "for a negative value of CC1/2. "
                           "(CC1/2 = " + str(bin_CChalf) + ")")
            else:
                try:
                    bin_CCstar = \
                        sqrt(2 * float(bin_CChalf) / (1 + float(bin_CChalf)))
                except ValueError:
                    bin_CCstar = "N/A"
                    warning_my("CC*", "A CC*-value for a particular "
                               "shell could not be calculated. "
                               "(CC1/2 = " + str(bin_CChalf) + ")")
            lines[i] = lines[i].replace("\n", "")
            lines[i] = lines[i] + "   " + fourdec(bin_CCstar) + "\n"
            # Insert number of the particular shell
            lines[i] = str(shell).zfill(2) + "    " + lines[i]
            shell = shell + 1
        return lines

    # Insert statistics relating up to resolution res_init
    for i in range(len(bins_total) - 1):
        logfilename = project + "_merging_stats_" \
            "" + twodecname(bins_total[i + 1]) + "A.log"
        line = extract_from_file(logfilename,
                                 searched="Statistics by resolution bin:",
                                 skip_lines=2, n_lines=1)
        os.remove(logfilename)  # Little clean-up
        line = calculate_CCstar(line, shell=i + 1)
        with open(csvfilename, "a") as csvfile:
            csvfile.writelines(line)
    return csvfilename


def run_baverage(project, xyzin, res_init):
    """Finds average value of B-factors of all the atoms in the structure
    model `xyzin` using `baverage` from the CCP4 package.
    This creates files with a prefix `project_twodecname(res_init)A_baverage`.
    Then returns mean B-factor for all the atoms of the structure model to
    the obtained value.

    Args:
        project (str): Name of the project
        xyzin (str): Filename of structure model in PDB or mmCIF format
        res_init (float): Resolution of the input structure model

    Returns:
        float: Mean B-factor for all the atoms
    """
    prefix = project + "_" + twodecname(res_init) + "A_baverage"
    logout = prefix + ".log"
    if "cif" in settings["pdbORmmcif"]:
        xyzout = prefix + settings["pdbORmmcif"]
    else:  # pdb
        xyzout = prefix + settings["pdbORmmcif"]
    command = ["baverage", "XYZIN", xyzin, "RMSTAB",
               prefix + ".tab", "XYZOUT", xyzout]
    com = "end\n"
    with open(logout, "w") as logfile:
        p = subprocess.Popen(command, stdin=subprocess.PIPE,
                             stdout=logfile, shell=settings["sh"])
        #               encoding='utf8')  # Probably required in Python 3
        p.communicate(com)

    baverage = extract_from_file(filename=logout,
                                 searched="AVERAGE B VALUE FOR ALL ATOMS",
                                 skip_lines=0,
                                 n_lines=1,
                                 nth_word=-1)
    baverage = float(baverage)
    print("Average B-factor for all atoms: " + twodec(baverage))
    # if add_to_bfactor:
        # baverage = str(float(baverage) + add_to_bfactor)
    #     bfac_set = baverage + add_to_bfactor
    #     print("B-factor after application of the parameter "
    #           "--add-to-bfactor: " + twodec(bfac_set))
    # else:
    #     bfac_set = baverage
    # bfac_set = baverage

    # Seems that it is not working well for anisotropic B-factors
    # http://www.ccp4.ac.uk/html/baverage.html
    # BLIMIT <bminmc> <bmaxmc> <bminsc> <bmaxsc>
    # prefix = project + "_" + twodecname(res_init) + "A_baveraged"
    # logout = prefix + ".log"
    # xyzout = prefix + ".pdb"
    # com = "BLIMIT " + baverage + " " + baverage + " " + baverage + " " \
    #     "" + baverage + "\nend\n"
    # command = ["baverage", "XYZIN", xyzin, "RMSTAB",
    #            prefix + ".tab", "XYZOUT", xyzout]
    # with open(logout, "w") as logfile:
    #     p = subprocess.Popen(command, stdin=subprocess.PIPE,
    #                          stdout=logfile)
    #     #               encoding='utf8')  # Probably required in Python 3
    #    p.communicate(com)
    # return xyzout
    # return bfac_set
    return baverage


def run_bmean_iotbx(project, xyzin):
    """Finds average value of B-factors of all the atoms in the structure
    model `xyzin` using `pdb` from `iotbx`.
    Then returns mean B-factor for all the atoms of the structure model to
    the obtained value.

    Args:
        project (str): Name of the project
        xyzin (str): Filename of structure model in PDB or mmCIF format

    Returns:
        float: Mean B-factor for all the atoms
    """
    from iotbx import pdb
    pdb_inp = pdb.input(file_name=xyzin)
    atoms = pdb_inp.atoms()
    bfactors = atoms.extract_b()
    baverage = float(bfactors.format_mean("%5.2f"))
    return baverage


def run_pdbtools(args, baverage=0):
    """Modify the input structure model `args.xyzin` by `mmtbx.pdbtools`.
    The procces is controlled by `args.reset_bfactor`, `args.add_to_bfactor`,
    `args.set_bfactor`, and `args.shake_sites`. Reset of B-factors requires
    set of `baverage`. The modified structure model:
    `project_twodecname(res_init)A_modified`.

    Args:
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()
        baverage (float): Mean B-factor for all the atoms

    Returns:
        str: Filename of the modified structure model (PDB or mmCIF format)
    """
    import shutil
    if not (args.reset_bfactor or args.add_to_bfactor or args.set_bfactor or
            args.shake_sites):
        # Nothing to do
        return args.xyzin

    prefix_phil = args.project + "_" + twodecname(args.res_init) + "A"
    prefix = args.project + "_" + twodecname(args.res_init) + "A_modified"
    logout = prefix + ".log"
    if "cif" in settings["pdbORmmcif"]:
        xyzout = prefix + ".cif"
    else:
        xyzout = prefix + ".pdb"

    phil_import_successful = None
    try:  # new versions of CCP4 or PHENIX
        from libtbx import phil
        from mmtbx.programs import pdbtools
        from iotbx.data_manager import DataManager
        phil_import_successful = True
    except ImportError:
        pass
    if phil_import_successful:
        dm = DataManager()
        dm.set_overwrite(True)
        model = dm.get_model(args.xyzin)
        phil_master = phil.parse(pdbtools.Program.master_phil_str,
                                 process_includes=True)
        obj_work = phil_master.extract()
        adp_obj = phil_master.extract().modify.adp[0]
        # prepare arguments
        if args.reset_bfactor and baverage:
            bfactor = baverage
            if args.add_to_bfactor:
                bfactor += float(args.add_to_bfactor)
                print("B-factor after application of the parameter "
                      "--prerefinement-add-to-bfactor: " + twodec(bfactor))
            adp_obj.set_b_iso = float(bfactor)
            # pdbtools_args.append("set_b_iso=" + twodec(bfactor))
        elif args.set_bfactor:
            adp_obj.set_b_iso = float(args.set_bfactor)
            # pdbtools_args.append("set_b_iso=" + twodec(args.set_bfactor))
        if args.add_to_bfactor and not args.reset_bfactor:
            adp_obj.shift_b_iso = float(args.add_to_bfactor)
            # pdbtools_args.append("shift_b_iso=" + twodec(args.add_to_bfactor))
        if adp_obj.set_b_iso or adp_obj.shift_b_iso:
            obj_work.modify.adp.append(adp_obj)
        if args.shake_sites:
            sites_obj = phil_master.extract().modify.sites[0]
            sites_obj.shake = float(args.shake_sites)
            obj_work.modify.sites.append(sites_obj)
            # pdbtools_args.append("shake=" + twodec(args.shake_sites))
        obj_work.output.prefix = prefix_phil
        obj_work.output.suffix = "_modified"
        return_phil = phil_master.format(obj_work)
        print("Modification of the input structure model using pdbtools...")
        Pdbtools = pdbtools.Program(data_manager=dm, params=obj_work)
        Pdbtools.run()
        return xyzout

    else:  #   old versions of CCP4 or PHENIX (no phil)
        pdbtools_args = []
        if args.reset_bfactor and baverage:
            bfactor = baverage
            if args.add_to_bfactor:
                bfactor += args.add_to_bfactor
                print("B-factor after application of the parameter "
                      "--prerefinement-add-to-bfactor: " + twodec(bfactor))
            pdbtools_args.append("set_b_iso=" + twodec(bfactor))
        elif args.set_bfactor:
            pdbtools_args.append("set_b_iso=" + twodec(args.set_bfactor))
        if args.add_to_bfactor and not args.reset_bfactor:
            pdbtools_args.append("shift_b_iso=" + twodec(args.add_to_bfactor))
        if args.shake_sites:
            pdbtools_args.append("shake=" + twodec(args.shake_sites))

        pdbtools_args.append("file_name=" + xyzout)
        print("Modification of the input structure model - pdbtools arguments: " + 
            " ".join(pdbtools_args))
        if which("phenix.pdbtools"):
            with open(logout, "w") as logfile:
                p = subprocess.Popen(["phenix.pdbtools", args.xyzin] + pdbtools_args,
                                     stdout=logfile, shell=settings["sh"])
                #               encoding='utf8')  # Probably required in Python 3
                p.communicate()
        else:
            import mmtbx.command_line.pdbtools
            pdbtools_args.append("model_file_name=" + args.xyzin)
            with open(logout, "w") as logfile:
                mmtbx.command_line.pdbtools.run(pdbtools_args, out=logfile,
                                                replace_stderr=False)
    
    if os.path.isfile(xyzout):
        if "cif" in settings["pdbORmmcif"]:
            # refmac required .mmcif (.cif does not work)
            shutil.copy2(xyzout, xyzout[:-4] + ".mmcif")
            xyzout = xyzout[:-4] + ".mmcif"
        return xyzout
    else:
        sys.stderr.write("ERROR: File " + xyzout + " has not been created "
                         "by pdbtools. Check the log file " + logout + "."
                         "\nAborting.\n")
        sys.exit(1)



import pytest
import os
import shlex
import subprocess
import sys
import tempfile


MODULE = "pairef"


def run(line, **kwargs):
    # try:
    #      if os.environ["TRAVIS"] == "true":
    #         # line += (" -t")
    #         # print("Light-testing for TRAVIS CI - not executing REFMAC5")
    #         pass
    # except KeyError:
    #     # print("Normal testing mode.")
    #     pass
    print('\n$ cctbx.python -m pairef', line)
    # print('\n$ python -m', MODULE, line)
    command = ["cctbx.python", '-m', "pairef"] + shlex.split(line)
    # return subprocess.run(command,
    process = subprocess.Popen(command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               universal_newlines=True,
                               **kwargs)
    cp = AttrDict()
    cp.stdout, cp.stderr = process.communicate()
    cp.returncode = process.returncode
    return cp


def run_ok(*args, **kwargs):
    cp = run(*args, **kwargs)
    assert cp.returncode == 0
    assert not cp.stderr
    print(cp.stdout)
    return cp


def config(name):
    return os.path.dirname(os.path.realpath(__file__)) + "/fixtures/" + name
    # return pathlib.Path(__file__).parent / 'fixtures' / name


@pytest.fixture
def tmp_environ():
    os.chdir(tempfile.gettempdir())
    return


# https://stackoverflow.com/
# questions/4984647/accessing-dict-keys-like-an-attribute
class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self



# coding: utf-8
from __future__ import print_function
from __future__ import division
import os
import sys
import re
import subprocess
import shutil
from math import sqrt
from .settings import warning_dict, settings
from .commons import twodec, twodecname, fourdec, extract_from_file, warning_my


def refinement_refmac(res_cur,
                      res_prev,
                      res_high,
                      args,
                      n_bins_low,
                      mode="refine",
                      res_low=0,
                      res_highest=0,
                      flag=0,
                      xyzin_start="",
                      bfac_set=0):  # TODO: hklfree
    """Refine using REFMAC5.

    Resolution limit is controlled by parameters `res_high` and `res_low`.
    Names of files are controlled by parameters `res_cur` and `res_high`.

    There are 4 different modes:

    * `mode="refine"`

          refine structure `args.project_twodecname(res_prev)A.pdb`;
          results with prefix `args.project_twodecname(res_cur)A`;
          `ncyc` given by `comfile` (default: `ncyc 10`)

    * `mode="comp"`

          use structure `args.project_twodecname(res_cur)A.pdb`;
          results with prefix `args.project_twodecname(res_cur)
          A_comparison_at_twodecname(res_high)A`;
          `ncyc 0`

    * `mode="prev_pair"`

          use structure `args.project_twodecname(res_cur)A.pdb`;
          results with prefix `args.project_twodecname(res_cur)
          "A_comparison_at_twodecname(res_high)A_prev_pair`;
          `ncyc 0`

    * `mode="first"`

          use structure `xyzin_start`);
          results with prefix `structure args.project_twodecname(res_cur)A.pdb`
          plus extra copy of log `structure args.project_
          twodecname(res_cur)A_comparison_at_twodecname(res_high)A.log`;
          `ncyc` controlled by an option `--prerefinement-ncyc` (0 cycles
          by default, 20 cycles by default for the complete
          cross-validation);
          usually `res_cur = res_init`, `res_high = res_init`

    If the input structure model was in mmCIF format, files with suffix .mmcif
    (not .pdb) are used.

    Args:
        res_cur (float)
        res_prev (float or str)
        res_high (float)
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()
        n_bins_low (int)
        mode (str): [`"refine"`, `"comp"`, `"prev_pair"`, `"first"`]
        res_low (float)
        res_highest (float)
        flag (int)
        xyzin_start (str): Filename of the model to be refined (valid only \
                           for `mode="first"`)
        bfac_set (float): Value of B-factor that will be set to all atoms
                          before refinement (not used now)

    Returns:
        (dict):
            Dictionary containing names of files that have been created
            by REFMAC5 and a version of REFMAC5, *e. i.*
            `HKLOUT`, `XYZOUT`, `LOGOUT`, and `version` (all `str`)
    """
    if res_low:
        reso = twodec(res_low) + " " + twodec(res_high)
    else:
        reso = twodec(res_high)
        res_low = "Dmax"

    if not args.comin:
        com = """
make -
    check NONE
refi -
    type REST reso """ + reso + """ -
    resi MLKF -
    meth CGMAT -
    bref MIXED
scal -
    type SIMP -
    LSSC -
    ANISO -
    EXPE
solvent YES
"""
    else:                                 # args.comin is str, name of the file
        with open(args.comin, "r") as comfile:  # comfile is the opened file
            com = comfile.read()   # com is str (modified content of the file)
        re_end = re.compile(re.escape("end"), re.IGNORECASE)
        com = re_end.sub("", com)
        # com = re.sub("(?i)end","", com) # case-insensitive
        com += "\n refi reso " + reso + " \n"
        if not any(["ncyc" in line.lower() for line in com.splitlines()]) \
           and not args.ncyc:
            args.ncyc = 10
    prefix = args.project + "_R" + str(flag).zfill(2) + "_" \
        "" + twodecname(res_cur) + "A"
    if mode == "comp":
        print(" .", end="")
        # print("       Calculating statistics of the refined structure model."
        ncyc_not_zero = False
        prefix += "_comparison" \
            "_at_" + twodecname(res_high) + "A"
        com += "\n ncyc 0"
        com += "\n bins " + str(n_bins_low)
    elif mode == "prev_pair":
        if args.complete_cross_validation:
            print(" .")
        else:
            print(" .", end="")
        ncyc_not_zero = False
        prefix += "_comparison_at_" + twodecname(res_high) + "A_prev_pair"
        com += "\n ncyc 0"
        com += "\n bins " + str(n_bins_low)
    elif mode == "first":
        prefix += ""
        if args.complete_cross_validation:
            ncyc_not_zero = True
            if args.prerefinement_ncyc:
                com += "\n ncyc " + str(args.prerefinement_ncyc)
            elif args.ncyc:
                com += "\n ncyc " + str(args.ncyc)
            elif not args.comin:
                com += "\n ncyc 20"
        else:
            if args.prerefinement_ncyc:
                ncyc_not_zero = True
                com += "\n ncyc " + str(args.prerefinement_ncyc)
            else:
                ncyc_not_zero = False
                com += "\n ncyc 0"
        if args.quick:
            ncyc_not_zero = True
            com += "\n ncyc 1"
        com += "\n bins " + str(n_bins_low)
        xyzin = xyzin_start
    elif mode == "refine":
        prefix += ""
        com += "\n bins " + str(n_bins_low)
        ncyc_not_zero = True
        if args.ncyc:
            com += "\n ncyc " + str(args.ncyc)
        elif not args.comin:
            com += "\n ncyc 10"
        if args.quick:
            com += "\n ncyc 1"
    if bfac_set:
        com += "\n BFACtor SET " + twodec(bfac_set)
    if args.weight:
        com += "\n weight matrix " + fourdec(args.weight)
    # Shannon factor
    if args.constant_grid and res_highest:
        shannon_factor = 1.5 * res_high / res_highest
        com += "\n SHANnon_factor " + fourdec(shannon_factor)
    # TLS - TLSC line
    if args.tlsin and ncyc_not_zero:
        if args.tls_ncyc:
            com += "\n refi tlsc " + str(args.tls_ncyc)
        else:
            com += "\n refi tlsc 10"
    com += "\n monitor MEDIum"
    com += "\n free " + str(flag)
    com += "\n END"
    hklout = prefix + ".mtz"
    if "cif" in settings["pdbORmmcif"]:
        xyzout = prefix + settings["pdbORmmcif"]
        xyzout_secondary = prefix + ".pdb"
    else:  # pdb
        xyzout = prefix + settings["pdbORmmcif"]
        xyzout_secondary = prefix + ".mmcif"
    libout = prefix + ".cif"
    logout = prefix + ".log"
    if mode == "refine":
        xyzin = args.project + "_R" + str(flag).zfill(2) + "_" + \
            twodecname(res_prev) + "A" + settings["pdbORmmcif"]
    elif mode == "comp" or mode == "prev_pair":
        xyzin = args.project + "_R" + str(flag).zfill(2) + "_" + \
            twodecname(res_cur) + "A" + settings["pdbORmmcif"]
    command = ["refmac5", "HKLIN", args.hklin, "XYZIN", xyzin, "HKLOUT",
               hklout, "XYZOUT", xyzout, "LIBOUT", libout]
    # Optional LIBIN
    if args.libin:
        command.append("LIBIN")
        command.append(args.libin)
    # Optional TLSIN (and TLSOUT)
    if args.tlsin:
        if ncyc_not_zero:
            command.append("TLSIN")
            if args.tlsin_keep:
                command.append(args.tlsin)
            else:
                if mode == "first":
                    tlsin = args.tlsin
                elif mode == "refine":
                    tlsin = args.project + "_R" + str(flag).zfill(2) + "_" + \
                        twodecname(res_prev) + "A.tlsout"
                command.append(tlsin)
            command.append("TLSOUT")
            tlsout = prefix + ".tlsout"
            command.append(tlsout)
        elif mode == "first":  # and ncyc 0
            shutil.copy2(args.tlsin, prefix + ".tlsout")

    if (mode == "refine" or
            (mode == "first" and args.complete_cross_validation)):
        if args.complete_cross_validation:
            print("     – FreeRflag set " + str(flag))
        print("       Running command:")
        print("       " + " ".join(command))
    with open(logout, "w") as logfile:
        p = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=logfile)
        #                    encoding='utf8')  # Probably required in Python 3
        p.communicate(com)
    for fileout in [logout, hklout, xyzout]:
        if not os.path.isfile(fileout):
            sys.stderr.write("ERROR: File " + fileout + " has not been created"
                             " by REFMAC5. Check a file " + logout + " for "
                             "the details.\nAborting.\n")
            sys.exit(1)
    # Copy log and tls while running REFMAC5 at the starting resolution
    if mode == "first":
        prefix_copy = prefix + "_comparison_at_" + twodecname(res_high) + "A"
        shutil.copy2(logout, prefix_copy + ".log")
        shutil.copy2(hklout, prefix_copy + ".mtz")
    version = extract_from_file(logout, "version", 0, 1, nth_word=5,
                                get_first=True)
    results = {"HKLOUT": hklout, "XYZOUT": xyzout, "LOGOUT": logout,
               "version": version}
    if mode == "comp" or mode == "prev_pair":
        files_to_be_removed = [xyzout, xyzout_secondary, libout]
        if "tlsout" in vars():
            to_be_removed += [tlsout]
        for filename in files_to_be_removed:
            if os.path.exists(filename):
                os.remove(filename)
    return results


def refinement_phenix(res_cur,
                      res_prev,
                      res_high,
                      args,
                      n_bins=1,
                      mode="refine",
                      res_low=0,
                      res_highest=0,
                      flag=0,
                      xyzin_start="",
                      bfac_set=0,
                      label=None):  # TODO: hklfree
    """Refine using phenix.refine.

    Resolution limit is controlled by parameters `res_high` and `res_low`.
    Names of files are controlled by parameters `res_cur` and `res_high`.

    There are 4 different modes:

    * `mode="refine"`

          refine structure `args.project_twodecname(res_prev)A_001.pdb`;
          results with prefix `args.project_twodecname(res_cur)A_001`;
          `number_of_macro_cycles` given by `deffile` (default: 3)

    * `mode="comp"`

          use structure `args.project_twodecname(res_cur)A_001.pdb`;
          results with prefix `args.project_twodecname(res_cur)
          A_comparison_at_twodecname(res_high)A_001`;
          `number_of_macro_cycles=1`, `refine.strategy=None`,
          ordered_solvent=False

    * `mode="prev_pair"`

          use structure `args.project_twodecname(res_cur)A_001.pdb`;
          results with prefix `args.project_twodecname(res_cur)
          "A_comparison_at_twodecname(res_high)A_prev_pair_001`;
          `number_of_macro_cycles=1`, `refine.strategy=None`,
          ordered_solvent=False

    * `mode="first"`

          use structure `xyzin_start`;
          results with prefix 
          `structure args.project_twodecname(res_cur)A_001.pdb`
          plus extra copy of log `structure args.project_
          twodecname(res_cur)A_comparison_at_twodecname(res_high)A_001.log`;
          `number_of_macro_cycles` controlled by an option 
          `--prerefinement-ncyc` (1 cycles with no strategy and no ordered
          solvent by default, 6 cycles by default for the complete
          cross-validation);
          usually `res_cur = res_init`, `res_high = res_init`

    If the input structure model was in mmCIF format, files with suffix .cif
    (not .pdb) are used.

    Args:
        res_cur (float)
        res_prev (float or str)
        res_high (float)
        args (parser): Input arguments (including e. g. name of the project) \
                       parsed by `argparse` via function process_arguments()
        n_bins (int)
        mode (str): [`"refine"`, `"comp"`, `"prev_pair"`, `"first"`]
        res_low (float)
        res_highest (float)
        flag (int)
        xyzin_start (str): Filename of the model to be refined (valid only \
                           for `mode="first"`)
        bfac_set (float): Value of B-factor that will be set to all atoms
                          before refinement (not used now)

    Returns:
        (dict):
            Dictionary containing names of files that have been created
            by phenix.refine and a version of phenix.refine, *e. i.*
            `HKLOUT`, `XYZOUT`, `LOGOUT`, and `version` (all `str`)
    """
    prefix = args.project + "_R" + str(flag).zfill(2) + "_" \
        "" + twodecname(res_cur) + "A"
    com = "refinement.input.xray_data.high_resolution=" + twodec(res_high)
    if res_low:
        com += "\nrefinement.input.xray_data.low_resolution=" + twodec(res_low)
    if args.defin:
        with open(args.defin, "r") as deffile:
            deff = deffile.read()
            if not any(["number_of_macro_cycles" in line for line in deff.splitlines()]) \
               and not args.ncyc:
                args.ncyc = 3

    if mode == "comp":
        print(" .", end="")
        # print("       Calculating statistics of the refined structure model."
        ncyc_not_zero = False
        prefix += "_comparison" \
            "_at_" + twodecname(res_high) + "A"
        com += "\nrefinement.main.number_of_macro_cycles=1"
        com += "\nrefinement.refine.strategy=None"
        com += "\nrefinement.main.ordered_solvent=False"
    elif mode == "prev_pair":
        if args.complete_cross_validation:
            print(" .")
        else:
            print(" .", end="")
        ncyc_not_zero = False
        prefix += "_comparison_at_" + twodecname(res_high) + "A_prev_pair"
        com += "\nrefinement.main.number_of_macro_cycles=1"
        com += "\nrefinement.refine.strategy=None"
        com += "\nrefinement.main.ordered_solvent=False"
    elif mode == "first":
        prefix += ""
        if args.complete_cross_validation:
            ncyc_not_zero = True
            if args.prerefinement_ncyc:
                com += "\nrefinement.main.number_of_macro_cycles=" + \
                    str(args.prerefinement_ncyc)
            elif args.ncyc:
                com += "\nrefinement.main.number_of_macro_cycles=" + str(args.ncyc)
            elif not args.comin:
                com += "\nrefinement.main.number_of_macro_cycles=6"
        else:
            if args.prerefinement_ncyc:
                ncyc_not_zero = True
                com += "\nrefinement.main.number_of_macro_cycles=" + \
                    str(args.prerefinement_ncyc)
            else:
                ncyc_not_zero = False
                com += "\nrefinement.main.number_of_macro_cycles=1"
                com += "\nrefinement.refine.strategy=None"
                com += "\nrefinement.main.ordered_solvent=False"
        if args.quick:
            ncyc_not_zero = True
            com += "\nrefinement.main.number_of_macro_cycles=1"
        xyzin = xyzin_start
    elif mode == "refine":
        prefix += ""
        ncyc_not_zero = True
        if args.ncyc:
            com += "\nrefinement.main.number_of_macro_cycles=" + str(args.ncyc)
        elif not args.comin:
            com += "\nrefinement.main.number_of_macro_cycles=3"
        if args.quick:
            com += "\nrefinement.main.number_of_macro_cycles=1"
    com += "\nrefinement.output.n_resolution_bins=" + str(n_bins)
    com += "\nrefinement.output.prefix=" + prefix
    com += "\nrefinement.input.xray_data.r_free_flags.test_flag_value=" + \
        str(flag)
    if hasattr(args, "label"):  # always false for the very first mode="refine"
        com += "\nrefinement.input.xray_data.labels=" + args.label
    # Output filenames
    if "cif" in settings["pdbORmmcif"]:
        xyzout = prefix + "_001" + settings["pdbORmmcif"]
        xyzout_secondary = prefix + "_001" + ".pdb"
    else:  # pdb
        xyzout = prefix + "_001" + settings["pdbORmmcif"]
        xyzout_secondary = prefix + "_001" + ".cif"
    hklout = prefix + "_001.mtz"
    logout = prefix + "_001.log"
    geoout = prefix + "_001.geo"
    outout = prefix + "_001.out"
    params = prefix + "_001.params"
    if mode == "refine":
        xyzin = args.project + "_R" + str(flag).zfill(2) + "_" + \
            twodecname(res_prev) + "A_001" + settings["pdbORmmcif"]
    elif mode == "comp" or mode == "prev_pair":
        xyzin = args.project + "_R" + str(flag).zfill(2) + "_" + \
            twodecname(res_cur) + "A_001" + settings["pdbORmmcif"]
    command = ["phenix.refine", args.hklin, xyzin]
    if args.libin:
        command.append(args.libin)
    if args.defin:
        command.append(args.defin)
    with open(params, "w") as pars:
        pars.write(com)
    command.append(params)

    if (mode == "refine" or
            (mode == "first" and args.complete_cross_validation)):
        if args.complete_cross_validation:
            print("     – FreeRflag set " + str(flag))
        print("       Running command:")
        print("       " + " ".join(command))
    with open(outout, 'w') as out:
        p = subprocess.Popen(command, stdout=out, stderr=out,
                             shell=settings["sh"])
        #                    encoding='utf8')  # Probably required in Python 3
        p.communicate()
    for fileout in [logout, hklout, xyzout, outout]:
        if not os.path.isfile(fileout):
            error = True
            if mode == "first" and not hasattr(args, "label"):
                label, labels_all = refinement_phenix_get_label(outout)
                if label:
                    warning_my(
                        "amb_labels", "Multiple equally suitable arrays of "
                        "observed xray data found. Possible choices: " + \
                        labels_all + " . Automatically choosing "
                        "refinement.input.xray_data.labels=" + label)
                    args.label = label  # a bit dirty hack but it works
                    results = refinement_phenix(res_cur=res_cur,
                                                res_prev=res_prev,
                                                res_high=res_high,
                                                args=args,
                                                n_bins=n_bins,
                                                mode="first",
                                                res_low=res_low,
                                                res_highest=res_highest,
                                                flag=flag,
                                                xyzin_start=xyzin_start)
                    error = False
            if error:
                sys.stderr.write("ERROR: File " + fileout + " has not been "
                                 "created by phenix.refine. Check a file " + \
                                 outout + " for the details."
                                 "\nAborting.\n")
                sys.exit(1)
    # Copy log and tls while running phenix.refine at the starting resolution
    if mode == "first":
        pdbout = prefix + "_001" + ".pdb"
        prefix_copy = prefix + "_comparison_at_" + twodecname(res_high) + "A"
        shutil.copy2(pdbout, prefix_copy + "_001.pdb")
    version = extract_from_file(logout, "Version", 0, 1, nth_word=1,
                                get_first=True)
    results = {"HKLOUT": hklout, "XYZOUT": xyzout, "LOGOUT": logout,
               "version": version}
    if mode == "comp" or mode == "prev_pair":
        files_to_be_removed = [hklout, geoout]
        if "tlsout" in vars():
            to_be_removed += [tlsout]
        for filename in files_to_be_removed:
            if os.path.exists(filename):
                os.remove(filename)
    return results


def refinement_phenix_get_label(outout):
    """Get possible choices of refinement.input.xray_data.labels from 
    standard output from phenix.refine (saved in file `outout`) if
    multiple equally suitable arrays of observed xray data were found.

    Args:
        out (str): Filename

    Returns:
        (tuple): tuple containing the chosen label and all the found labels - \
                 both are (`str`)"""
    signal = extract_from_file(
        filename=outout,
        searched="Multiple equally suitable arrays of observed xray data foun",
        skip_lines=0, n_lines=0, not_found="N/A")
    if signal == ["N/A"]:
        return None, None
    with open(outout, "r") as o:
        lines = o.read().splitlines()
    for i in range(len(lines)):
        if "Possible choices:" in lines[i]:
            j = i
            choices = []
    if "j" not in vars():
        return None, None
    while lines[j] != "":
        j += 1
        choices.append(lines[j].strip())
    labels_all = " ".join(choices)
    for choice in reversed(choices):
        if "+" in choice and len(choices) > 1:
            choices.remove(choice)
    label = choices[0]
    return label, labels_all


def collect_stat_BINNED(shells, project, hklin, n_bins_low, flag,
                        res_low, refinement="refmac"):
    """Collects statistics of a particular structure model depending on
    resolution (*e. i.* values are binned) and saves them in a CSV file.
    Statistics are picked from a REFMAC5 logfile relating to the model
    `project_RXX_twodecname(shells[-1])A.pdb`, where `XX` is a number
    of a flag.

    This function is called by the function `main()` in file `launcher.py`.

    Args:
        shells (list) - Contains high resolution diffraction limits
                        from the initial to the last which were used
                        for that model (float).
        project (str): Name of the project
        hklin (str): Name of diffraction data MTZ file
        n_bins_low (int)
        flag (int)
        res_low (float): Low resolution of diffraction data `hklin`
                         (required for calling the function
                         `collect_stat_refmac_log_low()`)
        refinement (str): "refmac" or "phenix"

    Returns:
        str: Name of the created CSV file
    """
    print("       Collecting statistics from logfiles...")
    prefix = project + "_R" + str(flag).zfill(2) + "_" \
        "" + twodecname(shells[-1]) + "A"
    if refinement == "refmac":
        # Pick statistics for `n_bins_low` res. shells up to initial diffr. limit
        logfilename = prefix + "_comparison" \
            "_at_" + twodecname(shells[0]) + "A.log"
        mtzfilename = prefix + ".mtz"
        bin_res_mean, bin_res_low, bin_res_high, bin_Nwork, bin_Nfree, \
            bin_Rwork, bin_Rfree, bin_CCwork, bin_CCfree = \
            collect_stat_binned_refmac_low(
                logfilename, mtzfilename, hklin, n_bins_low, res_low, flag)
        # Pick overall values for data up to previous diffraction limit
        # (to be comparable pairwisely)
        filename_prefix = prefix
        overall_Rwork, overall_Rfree = \
            collect_stat_overall_refmac(prefix + ".log", flag)
        overall_CCwork, overall_CCfree = calculate_correlation(mtzfilename,
                                                               hklin, flag)
        overall_CCwork = fourdec(overall_CCwork)
        overall_CCfree = fourdec(overall_CCfree)
    elif refinement == "phenix":
        # Pick statistics for `n_bins_low` res. shells up to initial diffr. limit
        pdbfilename = prefix + "_comparison" \
            "_at_" + twodecname(shells[0]) + "A_001.pdb"
        bin_res_low, bin_res_high, bin_Nwork, bin_Nfree, \
            bin_Rwork, bin_Rfree, bin_CCwork, bin_CCfree = \
            collect_stat_binned_phenix_low(pdbfilename, n_bins_low, res_low)
        # Pick overall values for data up to previous diffraction limit
        # (to be comparable pairwisely)
        pdbfilename = prefix + "_001.pdb"
        overall_Rwork, overall_Rfree = \
            collect_stat_overall_phenix(pdbfilename)
        overall_CCwork = "N/A"
        overall_CCfree = "N/A"
        overall_CCavg = "N/A"
    # Write the statistics to a csv file
    csvfilename = prefix + ".csv"
    with open(csvfilename, "w") as csvfile:
        csvfile.write("# Statistics of refined structure model calculated "
                      "at resolution range " + twodec(res_low) + "-"
                      "" + twodec(shells[-1]) + " A.\n")
        csvfile.write("# Rwork = " + overall_Rwork + " \n")
        csvfile.write("# Rfree = " + overall_Rfree + " \n")
        csvfile.write("# CCwork = " + overall_CCwork + " \n")
        csvfile.write("# CCfree = " + overall_CCfree + " \n")
        csvfile.write("# CCavg = N/A \n")
        csvfile.write("#\n")
        csvfile.write("#Shell Res_low - Res_hig      Nwork    Nfree  Rwork    "
                      "Rfree   CCwork  CCfree \n")
    collect_stat_write(csvfilename, bin_res_low, bin_res_high,
                              bin_Nwork, bin_Nfree,
                              bin_Rwork, bin_Rfree,
                              bin_CCwork, bin_CCfree)
    # Pick stats. for the high resolution shells and write them to the CSV file
    for i in range(len(shells) - 1):
        if refinement == "refmac":
            bin_res_low = [twodec(shells[i])]
            bin_res_high = [twodec(shells[i + 1])]
            logfilename = prefix + "_comparison" \
                "_at_" + twodecname(shells[i + 1]) + "A.log"
            mtzfilename = prefix + ".mtz"
            bin_Nwork, bin_Nfree, bin_Rwork, bin_Rfree, bin_CCwork, \
                bin_CCfree = collect_stat_binned_refmac_high(
                    logfilename, mtzfilename, hklin, n_bins_low,
                    float(twodec(shells[i])), float(twodec(shells[i + 1])), flag)
        elif refinement == "phenix":
            pdbfilename = prefix + "_comparison" \
                "_at_" + twodecname(shells[i + 1]) + "A_001.pdb"
            bin_res_low, bin_res_high, bin_Nwork, bin_Nfree, bin_Rwork, bin_Rfree, \
                bin_CCwork, bin_CCfree = collect_stat_binned_phenix_high(pdbfilename, n_bins_low)
        collect_stat_write(
            csvfilename, bin_res_low, bin_res_high, bin_Nwork, bin_Nfree,
            bin_Rwork, bin_Rfree, bin_CCwork, bin_CCfree,
            shell_number=n_bins_low + 1 + i)
    return csvfilename


def collect_stat_binned_refmac_low(logfilename, mtzfilename, hklin,
                                   n_bins_low, res_low=999, flag=0):
    """Picks and returns statistics values in the given `REFMAC5` logfile.
    Logfile supposed to contain information from `n_bins_low` shells.

    This function is called by the function `collect_stat_BINNED()`
    from this file and `main()` from file `launcher.py`.

    Args:
        logfilename (str): Name of a `REFMAC5` logfile
        n_bins_low (int): Number of low resolution bins

    Returns:
        (tuple): tuple containing statistics
        values `bin_Nwork`, `bin_Nfree`, `bin_Rwork`,
        `bin_Rfree`, `bin_CCwork`, and `bin_CCfree` (all are `str`)
    """
    bin_4SSQLL_mean = []
    bin_res_mean = []
    bin_res_low = []
    bin_res_high = []
    bin_Nwork = []
    bin_Nfree = []
    bin_Rwork = []
    bin_Rfree = []
    bin_CCwork = []
    bin_CCfree = []

    # # Definition of bins, numbers of reflections, R-values
    logfile_lines = extract_from_file(
        logfilename, "Things for loggraph, R factor and others  ",
        11, n_bins_low)
    # Refmac gives only mean 4SSQLL, so res_high and res_low (in angtroem)
    # calculations are neccessary
    shell_step_4SSQLL = \
        float(logfile_lines[1].split()[0]) - float(logfile_lines[0].split()[0])
    for i in range(n_bins_low):
        bin_4SSQLL_mean.append(logfile_lines[i].split()[0])
        bin_res_mean.append(str(twodec(sqrt(1 / float(bin_4SSQLL_mean[i])))))
        if i == 0:
            # This is more accurate and there is no danger of zero division
            res_low_cur = res_low
        else:
            res_low_cur = sqrt(
                1 / (abs(float(bin_4SSQLL_mean[i]) - shell_step_4SSQLL / 2)))
        bin_res_low.append(twodec(res_low_cur))
        res_high_cur = sqrt(
            1 / (float(bin_4SSQLL_mean[i]) + shell_step_4SSQLL / 2))            
        bin_res_high.append(twodec(res_high_cur))
        bin_Nwork.append(logfile_lines[i].split()[1])
        try:
            bin_Nfree.append(logfile_lines[i].split()[7])
            bin_Rwork.append(logfile_lines[i].split()[5])
            bin_Rfree.append(logfile_lines[i].split()[10])
        except IndexError:
            bin_Nfree.append("N/A")
            bin_Rwork.append("N/A")
            bin_Rfree.append("N/A")
            warning_my("low_R", "All R-values relating to the resolution"
                       " bin " + twodec(bin_res_low[-1]) + "-"
                       "" + twodec(bin_res_high[-1]) + " were not calculated "
                       "successfully. "
                       "For further details, see file " + logfilename + ".")
        CCwork, CCfree = calculate_correlation(
            mtzfilename, hklin, flag, res_low_cur, res_high_cur)
        bin_CCwork.append(fourdec(CCwork))
        bin_CCfree.append(fourdec(CCfree))
    return(bin_res_mean, bin_res_low, bin_res_high,
           bin_Nwork, bin_Nfree,
           bin_Rwork, bin_Rfree,
           bin_CCwork, bin_CCfree)


# TODO - what to do if CC-values are not available, strange log format
def collect_stat_binned_refmac_high(
    logfilename, mtzfilename, hklin, n_bins_low, res_low, res_high, flag=0):
    """Picks and returns statistics values in the given `REFMAC5` logfile.
    Logfile supposed to contain information from 1 shells.

    This function is called by the function `collect_stat_BINNED()`.

    Args:
        logfilename (str): Name of a `REFMAC5` logfile
        n_bins_low (int): Number of low resolution bins

    Returns:
        (tuple): tuple containing statistics
        values `bin_Nwork`, `bin_Nfree`, `bin_Rwork`,
        `bin_Rfree`, `bin_CCwork`, and `bin_CCfree` (all are `str`)
    """
    # R-values and CC-values
    bin_Rwork = [extract_from_file(logfilename, "Overall R factor", 0, 1, -1)]
    bin_Rfree = [extract_from_file(logfilename, "Free R factor", 0, 1, -1)]
    if "*" in bin_Rwork:
        bin_Rwork = "N/A"  # TODO: Seems it is not working now...
        bin_Rfree = "N/A"
        warning_my("high_R", "R-values of a particular shell "
                   "were not calculated. For "
                   "further details, see file " + logfilename + ".")
    CCwork, CCfree = calculate_correlation(
        mtzfilename, hklin, flag, res_low, res_high)
    bin_CCwork = [fourdec(CCwork)]
    bin_CCfree = [fourdec(CCfree)]
    bin_Nwork = [extract_from_file(logfilename, "Number of used reflections",
                 0, 1, -1)]

    # A number of reflections in bin
    bin_Nfree = []
    logfile_lines = extract_from_file(
        logfilename, "Things for loggraph, R factor and others  ",
        11, n_bins_low)
    try:
        for i in range(n_bins_low):
            bin_Nfree.append(int(logfile_lines[i][49:57]))
        bin_Nfree = [str(sum(bin_Nfree))]
    except IndexError:
        bin_Nfree = ["N/A"]
    return(bin_Nwork, bin_Nfree, bin_Rwork, bin_Rfree,
           bin_CCwork, bin_CCfree)


def collect_stat_binned_phenix_low(pdbfilename, n_bins_low, res_low=999):
    """Picks and returns statistics values in the given `REFMAC5` logfile.
    Logfile supposed to contain information from `n_bins_low` shells.

    This function is called by the function `collect_stat_BINNED()`
    from this file and `main()` from file `launcher.py`.

    Args:
        pdbfilename (str): Name of a `REFMAC5` logfile
        n_bins_low (int): Number of low resolution bins

    Returns:
        (tuple): tuple containing statistics
        values `bin_Nwork`, `bin_Nfree`, `bin_Rwork`,
        `bin_Rfree`, `bin_CCwork`, and `bin_CCfree` (all are `str`)
    """
    bin_res_low = []
    bin_res_high = []
    bin_Nwork = []
    bin_Nfree = []
    bin_Rwork = []
    bin_Rfree = []
    bin_CCwork = []
    bin_CCfree = []
    # # Definition of bins, numbers of reflections, R-values
    pdbfile_lines = extract_from_file(
        pdbfilename, "REMARK   3   BIN  RESOLUTION RANGE  COMPL.    "
        "NWORK NFREE   RWORK  RFREE  CCWORK CCFREE", 1, n_bins_low)
    for i in range(n_bins_low):
        bin_res_low.append(pdbfile_lines[i].split()[3])
        bin_res_high.append(pdbfile_lines[i].split()[5])
        bin_Nwork.append(pdbfile_lines[i].split()[7])
        bin_Nfree.append(pdbfile_lines[i].split()[8])
        bin_Rwork.append(pdbfile_lines[i].split()[9])
        bin_Rfree.append(pdbfile_lines[i].split()[10])
        bin_CCwork.append(pdbfile_lines[i].split()[11])
        bin_CCfree.append(pdbfile_lines[i].split()[12])
    return(bin_res_low, bin_res_high, bin_Nwork, bin_Nfree,
           bin_Rwork, bin_Rfree, bin_CCwork, bin_CCfree)


def collect_stat_binned_phenix_high(pdbfilename, n_bins_low):
    """Picks and returns statistics values in the given `REFMAC5` logfile.
    Logfile supposed to contain information from 1 shells.

    This function is called by the function `collect_stat_BINNED()`.

    Args:
        logfilename (str): Name of a `REFMAC5` logfile
        n_bins_low (int): Number of low resolution bins

    Returns:
        (tuple): tuple containing statistics
        values `bin_Nwork`, `bin_Nfree`, `bin_Rwork`,
        `bin_Rfree`, `bin_CCwork`, and `bin_CCfree` (all are `str`)
    """
    line = extract_from_file(
        pdbfilename, "REMARK   3   BIN  RESOLUTION RANGE  COMPL.    "
       "NWORK NFREE   RWORK  RFREE  CCWORK CCFREE", 1, 1, not_found="N/A")
    line = line[0]
    bin_res_low = [line.split()[3]]
    bin_res_high = [line.split()[5]]
    bin_Nwork = [line.split()[7]]
    bin_Nfree = [line.split()[8]]
    bin_Rwork = [line.split()[9]]
    bin_Rfree = [line.split()[10]]
    bin_CCwork = [line.split()[11]]
    bin_CCfree = [line.split()[12]]
    return(bin_res_low, bin_res_high, bin_Nwork, bin_Nfree,
           bin_Rwork, bin_Rfree, bin_CCwork, bin_CCfree)


def collect_stat_write(csvfilename, bin_res_low, bin_res_high,
                              bin_Nwork, bin_Nfree,
                              bin_Rwork, bin_Rfree,
                              bin_CCwork, bin_CCfree,
                              shell_number=1):
    """Saves the given statistics values to a CSV file.

    This function is called by the function `collect_stat_BINNED()`.

    Args:
        csvfilename (str): Filename of the CSV file to be created.
        bin_* (list): Lists with statistics values
        shell_number (int): Number of the shell corresponding to the
                            statistics values

    Returns:
        str: Filename of the created CSV file.
    """
    with open(csvfilename, "a") as csvfile:
        for i in range(len(bin_Rwork)):
            csvfile.write(
                str(shell_number).zfill(2) + "        " + bin_res_low[i] + ""
                " - " + bin_res_high[i] + "         " + bin_Nwork[i] + "     "
                "" + bin_Nfree[i] + "     " + bin_Rwork[i] + "     "
                "" + bin_Rfree[i] + "     " + bin_CCwork[i] + "     "
                "" + bin_CCfree[i] + "\n")
            if float(bin_Nfree[i]) < 50:
                warning_my("Nfree", "There are only " + str(bin_Nfree[i]) + ""
                           " < 50 free reflections in the resolution shell "
                           "" + twodec(bin_res_low[i]) + "-"
                           "" + twodec(bin_res_high[i]) + " A. Values of "
                           "statistics Rfree and CCfree in this shell "
                           "could be misleading. Consider setting "
                           "thicker resolution shells.")
            shell_number = shell_number + 1
    return csvfilename


def collect_stat_OVERALL(shells, args, flag, refinement="refmac"):
    """Collects overall statistics of a particular structure model
    from a REFMAC5 logfile or from PDB file (phenix.refine).
    Model which is dealing with: `project_RXX_twodecname(shells[-1])A.pdb`
    (REFMAC5) or `project_RXX_twodecname(shells[-1])A_001.pdb` (phenix.refine),
    where `XX` is a number of a flag.

    This function is called by the function `main()` in file `launcher.py`.

    Args:
        shells (list):  Contains high resolution diffraction limits
                        from the initial to the last which were used
                        for that model (float).
        project (str): Name of the project
        flag (int):
        refinement (str): "refmac" or "phenix"

    Returns:
        (tuple): tuple containing names of created CSV files
    """
    csvfilenames = []
    prefix = args.project + "_R" + str(flag).zfill(2)
    # Pick overall values - values calculated on the same data - "paired"
    if len(shells) >= 2:
        if refinement == "refmac":
            # get Rwork, Rfree, CCwork, CCfree, CCavg
            filename_prefix = prefix + "_" + twodecname(shells[-2]) + "A"
            Rwork_before, Rfree_before = \
                collect_stat_overall_refmac(filename_prefix + ".log", flag)
            CCwork_before, CCfree_before = calculate_correlation(
                filename_prefix + ".mtz", args.hklin, flag) # floats
            filename_prefix = prefix + "_" + twodecname(shells[-1]) + "A_" \
                "comparison_at_" + twodecname(shells[-2]) + "A_prev_pair"
            Rwork_after, Rfree_after = \
                collect_stat_overall_refmac(filename_prefix + ".log", flag)
            mtzfilename = prefix + "_" + twodecname(shells[-1]) + "A.mtz"
            CCwork_after, CCfree_after = calculate_correlation(
                mtzfilename, args.hklin, flag, res_high=float(shells[-2]))  # floats
        elif refinement == "phenix":
            # get Rwork, Rfree                (not CCwork, CCfree, CCavg)
            pdbfilename = prefix + "_" + twodecname(shells[-2]) + "A_001.pdb"
            Rwork_before, Rfree_before = \
                collect_stat_overall_phenix(pdbfilename)
            pdbfilename = prefix + "_" + twodecname(shells[-1]) + "A_comparison" \
                "_at_" + twodecname(shells[-2]) + "A_prev_pair_001.pdb"
            Rwork_after, Rfree_after = \
                collect_stat_overall_phenix(pdbfilename)

        # Rwork, Rfree
        Rwork_change = fourdec(float(Rwork_after) - float(Rwork_before))
        Rfree_change = fourdec(float(Rfree_after) - float(Rfree_before))
        #
        csvfilename = prefix + "_R-values.csv"
        csvfilenames.append(csvfilename)
        # Write the begining of the csv file if it does not exist yet
        if not os.path.isfile(csvfilename):
            with open(csvfilename, "w") as csvfile:
                csvfile.write("# Shell      Rwork(init) Rwork(fin) Rwork(diff)"
                              "   Rfree(init) Rfree(fin) Rfree(diff)\n")
        # Formating issues
        space_Rwork = ""
        space_Rfree = ""
        if float(Rwork_change) >= 0:
            space_Rwork = " "
        if float(Rfree_change) >= 0:
            space_Rfree = " "
        # Write values in the csv file
        with open(csvfilename, "a") as csvfile:
            csvfile.write(twodec(shells[-2]) + "A->" + twodec(shells[-1]) + "A"
                          "      " + Rwork_before + "     " + Rwork_after + ""
                          "     " + space_Rwork + Rwork_change + "        "
                          "" + Rfree_before + "     " + Rfree_after + "     "
                          "" + space_Rfree + Rfree_change + "\n")

    # Pick overall values of the current structure model
    # and find Rfree-Rwork gap (at the initial resolution)
    if refinement == "refmac":     # get Rwork, Rfree
        logfilename = prefix + "_" + twodecname(shells[-1]) + "A" \
            "_comparison_at_" + twodecname(shells[0]) + "A.log"
        Rwork, Rfree = collect_stat_overall_refmac(logfilename, flag)
    elif refinement == "phenix":   # get Rwork, Rfree
        pdbfilename = prefix + "_" + twodecname(shells[-1]) + "A" \
            "_comparison_at_" + twodecname(shells[0]) + "A_001.pdb"
        Rwork, Rfree = collect_stat_overall_phenix(pdbfilename)
    Rgap = fourdec(float(Rfree) - float(Rwork))
    csvfilename_gap = prefix + "_Rgap.csv"
    # Write the begining of the csv file if it does not exist yet
    if not os.path.isfile(csvfilename_gap):
        with open(csvfilename_gap, "w") as csvfile:
            csvfile.write("# Resolution   Rwork   Rfree   Rfree-Rwork\n")
    with open(csvfilename_gap, "a") as csvfile:
        csvfile.write(twodec(shells[-1]) + "          " + Rwork + "   "
                      "" + Rfree + "   " + Rgap + "\n")
    csvfilenames.append(csvfilename_gap)
    return tuple(csvfilenames)


def collect_stat_overall_refmac(logfilename, flag=0):
    """Picks and returns overall Rwork, Rfree from a given `REFMAC5` logfile.

    This function is called by the functions `collect_stat_OVERALL()`
    and `collect_stat_BINNED()`.

    Args:
        logfilename (str): Filename of a `REFMAC5` logfile

    Returns:
        (tuple): tuple containing statistics values `Rwork` and `Rfree` \
                 (both are `str`)
    """
    Rwork = extract_from_file(logfilename, "R factor", 0, 1, -1)
    Rfree = extract_from_file(logfilename, "R free", 0, 1, -1)
    if "*" in Rwork:
        Rwork = "N/A"
        Rfree = "N/A"
        warning_my("overall_R", "Some overall R-values were not calculated. "
                   "For further details, see file " + logfilename + ".")
    return Rwork, Rfree


def calculate_correlation(hkl_calc, hklin,
                          flag=0, res_low=None, res_high=None):
    """Calculates CCwork and CCfree using `sftools`.

    Args:
        hkl_calc (str): Name of the MTZ file from a REFMAC5 run
        hklin (str): Name of the MTZ file with diffraction data
        flag (int): free reflection flag set
        res_low (float): low-resolution cutoff
        res_high (float): high-resolution cutoff

    Returns:
        (tuple): tuple containing `CCwork` and `CCfree` \
                 (both are `float` or `str`: "N/A")
    """
    from iotbx.reflection_file_reader import any_reflection_file
    from scitbx.array_family import flex

    # i_obs from HKLIN
    i_obs = None
    miller_arrays = any_reflection_file(file_name=hklin).as_miller_arrays()
    for i, column in enumerate(miller_arrays):
        if column.is_xray_intensity_array() and \
                column.anomalous_flag() == False:
            i_obs_label = str(column.info()).split(".mtz:")[1].split(",")[0]
            i_obs = column
            break
    if not i_obs:
        print("Error - intensities were not found in " + hklin)  # TO DO
        return "N/A", "N/A"

    # prepare sftools command and run
    if res_low and res_high:  # else: full resolution of MTZ file
        com_resolution = " resolution " + fourdec(res_low) + " " + fourdec(res_high)
    elif res_high:
        com_resolution = " resolution 999 " + fourdec(res_high)
    else:
        com_resolution = ""
    p = subprocess.Popen(["sftools"], stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         shell=settings["sh"])
    com = "read " + hkl_calc + " col 1 FC_ALL\n" \
        "calc col FC_ALLsq = col FC_ALL col FC_ALL *\n" \
        "read " + hklin + " col " + i_obs_label + "\n" \
        "select only col 1 = " + str(flag) + "\n" \
        "correl col " + i_obs_label + " FC_ALLsq shells 1" + com_resolution + "\n" \
        "select only not col 1 = " + str(flag) + "\n" \
        "correl col " + i_obs_label + " FC_ALLsq shells 1" + com_resolution + "\n" \
        "stop\n" \
        "y\n"
    output, err = p.communicate(com)

    # pick values
    CCwork = None
    CCfree = None
    output_lines = output.splitlines()
    for i in range(len(output.splitlines())):
        if "WARNING" in output.splitlines()[i]:
            print(output.splitlines()[i])
            return "N/A", "N/A"
        elif " $TABLE: Correlation vs. resolution:" in output.splitlines()[i] and \
                "$$" in output.splitlines()[i + 2] and \
                "$$" in output.splitlines()[i + 4] and \
                "$$" in output.splitlines()[i + 6] and \
                "NSHELL  DMIN    DMAX    RFACT   CORREL    <F1>    <F2>    NREF $$" in output.splitlines()[i + 3]:
            if not CCfree:
                try:
                    CCfree = float(output.splitlines()[i + 5].split()[4])
                    Nfree = int(output.splitlines()[i + 5].split()[-1])
                except ValueError:
                    CCfree = "N/A"
                    Nfree = "N/A"
            else:  # CCfree already picked
                try:
                    CCwork = float(output.splitlines()[i + 5].split()[4])
                    Nwork = int(output.splitlines()[i + 5].split()[-1])
                except ValueError:
                    CCwork = "N/A"
                    Nwork = "N/A"
    if not CCfree:
        CCfree = "N/A"
    if not CCwork:
        CCwork = "N/A"
    return CCwork, CCfree

    
def collect_stat_overall_phenix(pdbfilename):
    """Picks and returns overall Rwork, Rfree values from a given PDB file 
    refined in phenix.refine.

    This function is called by the functions `collect_stat_OVERALL()`
    and `collect_stat_BINNED()`.

    Args:
        pdbfilename (str): Filename of a PDB structure model refined in \
                           phenix.refine

    Returns:
        (tuple): tuple containing statistics values `Rwork`, `Rfree` \
                 (all are `str`)
    """
    Rwork = extract_from_file(
        pdbfilename,
        "REMARK   3   R VALUE            (WORKING SET) : ", 0, 1, -1)
    Rfree = extract_from_file(
        pdbfilename,
        "REMARK   3   FREE R VALUE                     : ", 0, 1, -1)
    return Rwork, Rfree


def collect_stat_OVERALL_AVG(shells, project, flag_sets):
    """Calculates and saves average overall values from CSV files prepared by
    the function `collect_stat_OVERALL()`.

    This function is called by the function `main()` in file `launcher.py`.

    Args:
        shells (list): Contains high resolution diffraction limits
                       from the initial to the last which were used
                       for that model (float).
        project (str): Name of the project
        flag_sets (list): List of free reflection flag sets (int)

    Returns:
        (tuple): tuple containing names of created CSV files
    """
    import numpy as np
    # === Overall Rwork, Rfree ===
    Rwork_init = []
    Rwork_fin = []
    Rwork_diff = []
    Rfree_init = []
    Rfree_fin = []
    Rfree_diff = []
    for flag in flag_sets:
        csvfilename = project + "_R" + str(flag).zfill(2) + "_R-values.csv"
        with open(csvfilename, "r") as csvfile:
            csvfile_last_line = csvfile.readlines()[-1]
            Rwork_init.append(float(csvfile_last_line.split()[1]))
            Rwork_fin.append(float(csvfile_last_line.split()[2]))
            Rwork_diff.append(float(csvfile_last_line.split()[3]))
            Rfree_init.append(float(csvfile_last_line.split()[4]))
            Rfree_fin.append(float(csvfile_last_line.split()[5]))
            Rfree_diff.append(float(csvfile_last_line.split()[6]))
    Rwork_init_avg = fourdec(np.mean(Rwork_init))
    Rwork_fin_avg = fourdec(np.mean(Rwork_fin))
    Rwork_diff_avg = fourdec(np.mean(Rwork_diff))
    Rfree_init_avg = fourdec(np.mean(Rfree_init))
    Rfree_fin_avg = fourdec(np.mean(Rfree_fin))
    Rfree_diff_avg = fourdec(np.mean(Rfree_diff))
    Rwork_stdev = fourdec(np.std(Rwork_fin))
    Rfree_stdev = fourdec(np.std(Rfree_fin))
    Rwork_diff_sem = str(round(np.std(Rwork_diff)/sqrt(len(flag_sets)), 5))
    Rfree_diff_sem = str(round(np.std(Rfree_diff)/sqrt(len(flag_sets)), 5))
    # Rwork_diff_avg2 = fourdec(float(Rwork_fin_avg) - float(Rwork_init_avg))
    # Rfree_diff_avg2 = fourdec(float(Rfree_fin_avg) - float(Rfree_init_avg))

    csvfilename = project + "_R-values.csv"
    if not os.path.isfile(csvfilename):
        with open(csvfilename, "w") as csvfile:
            csvfile.write("# Shell      Rwork(init) Rwork(fin) Rwork(diff)"
                          "   Rfree(init) Rfree(fin) Rfree(diff)   "
                          "Rwork(StDev)   Rfree(StDev)  "
                          "Rwork(diff,SEM) Rfree(diff,SEM)\n")

    # Formating issues
    space_Rwork = ""
    space_Rfree = ""
    if float(Rwork_diff_avg) >= 0:
        space_Rwork = " "
    if float(Rfree_diff_avg) >= 0:
        space_Rfree = " "
    with open(csvfilename, "a") as csvfile:
        csvfile.write(twodec(shells[-2]) + "A->" + twodec(shells[-1]) + "A"
                      "      " + Rwork_init_avg + "     " + Rwork_fin_avg + ""
                      "     " + space_Rwork + Rwork_diff_avg + "        "
                      "" + Rfree_init_avg + "     " + Rfree_fin_avg + "     "
                      "" + space_Rfree + Rfree_diff_avg + "         "
                      "" + Rwork_stdev + "       " + Rfree_stdev + "     "
                      "" + Rwork_diff_sem + "       " + Rfree_diff_sem + "\n")

    # === Rgap ===
    Rgap_avg = fourdec(float(Rfree_fin_avg) - float(Rwork_fin_avg))
    csvfilename_gap = project + "_Rgap.csv"
    if not os.path.isfile(csvfilename_gap):
        with open(csvfilename_gap, "w") as csvfile:
            csvfile.write("# Resolution   Rwork   Rfree   Rfree-Rwork\n")
    with open(csvfilename_gap, "a") as csvfile:
        csvfile.write(twodec(shells[-1]) + "          " + Rwork_fin_avg + "   "
                      "" + Rfree_fin_avg + "   " + Rgap_avg + "\n")
    return csvfilename, csvfilename_gap



import pytest
import platform
import sys
import os
import tempfile
import shutil
from helper import run, config, tmp_environ
from pairef.preparation import create_workdir, which


def test_python_version():
    assert int(platform.python_version_tuple()[0]) == 2
    assert int(platform.python_version_tuple()[1]) >= 7


RES_SHELLS_GOOD = "1.9,1.8,1.7,1.6,1.5,1.4"

prog = 'cctbx.python -m pairef'
help = \
"""usage: """ + prog + """ --XYZIN XYZIN --HKLIN HKLIN [-u HKLIN_UNMERGED]
                              [--LIBIN LIBIN] [--TLSIN TLSIN] [-c COMIN]
                              [-p PROJECT] [-r RES_SHELLS] [-n N_SHELLS]
                              [-s STEP] [-i RES_INIT] [-f FLAG] [-w WEIGHT]
                              [--ncyc NCYC] [--constant-grid] [--complete]
                              [--TLS-ncyc TLS_NCYC] [--TLSIN-keep] [-h]
                              [--prerefinement-ncyc PREREFINEMENT_NCYC]
                              [--prerefinement-reset-bfactor]
                              [--prerefinement-add-to-bfactor ADD_TO_BFACTOR]
                              [--prerefinement-set-bfactor SET_BFACTOR]
                              [--prerefinement-shake-sites [SHAKE_SITES]]
                              [--prerefinement-no-modification]"""


def test_no_args(tmp_environ):
    cp = run("")
    # assert not cp.stdout
    # assert cp.stderr == help + "\n" + prog + \
    #     ": error: argument --XYZIN/--xyzin is required\n"
    assert cp.stderr == "error: argument --XYZIN/--xyzin is required\n"


def test_file_not_exists(tmp_environ):
    cp = run("--HKLIN " + str(config("foo.mtz")) + ""
             " --XYZIN " + str(config("lysozyme_arp_model_2A.pdb")) + ""
             " -i 2" + ""
             " -r " + RES_SHELLS_GOOD + ""
             " -p file_not_exists")
    assert cp.returncode != 0
    # assert not cp.stdout
    # assert cp.stderr == help + "\ncctbx.python -m pairef: error: " \
    #     "The file " + str(config("foo.mtz")) + " does not exist!\n"
    assert cp.stderr == "error: " \
        "The file " + str(config("foo.mtz")) + " does not exist!\n"


error_res_shells_not_floats = "ERROR: Explicit definition of high " \
    "resolution shells (option -r) is not correct. " \
    "Values must be divided using commas without " \
    "any spaces (e.g. 2.1,2.0,1.9).\nAborting.\n"
# error_res_init_not_float = help + "\n" + prog + \
#     ": error: argument -i: invalid float value: 'abc'\n"
error_res_init_not_float = "error: argument -i: invalid float value: 'abc'\n"
error_only_one_value = "ERROR: Explicit definition of high " \
    "resolution shells (option -r) is not correct. " \
    "Values must be divided using commas without " \
    "any spaces (e.g. 2.1,2.0,1.9). " \
    "More values (than only one) must be provided as the first " \
    "value is the conservative diffraction limit.\nAborting.\n"
error_invalid_order = "ERROR: Explicit definition of high " \
    "resolution shells (option -r) is not correct. " \
    "Values must be set in the decreasing order " \
    "(e.g. 2.1,2.0,1.9).\nAborting.\n"
error_invalid_res_init1 = "ERROR: Explicit definition of high " \
    "resolution shells (option -r) is not correct. " \
    "Resolution of the first shell (2.20 A), " \
    "which was explicitely defined, is lower than or the same as " \
    "the resolution of data which were used for refinement of the " \
    "input structure model (2.10 A).\nAborting.\n"
error_invalid_res_init2 = "ERROR: Explicit definition of high " \
    "resolution shells (option -r) is not correct. " \
    "Resolution of the first shell (2.10 A), " \
    "which was explicitely defined, is lower than or the same as " \
    "the resolution of data which were used for refinement of the " \
    "input structure model (2.10 A).\nAborting.\n"


# This is testing the fuction def_res_shells
@pytest.mark.parametrize(
    ["res_init", "res_shells", "error_message"],
    # # [  # #("2", "abc", error_res_shells_not_floats),
    [("abc", "2.2,2,1.8", error_res_init_not_float)],
    # ("2.0", error_message_only_one_value),
    # #("2.1", "1.9,3.5,1.8", error_invalid_order),
    # #("2.1", "2.2,2,1.9", error_invalid_res_init1),
    # #("2.1", "2.1,2.0,1.9", error_invalid_res_init2)],
    # # ],
    # # ids=["res_shells_not_floats",
    ids=["res_init_not_float"])
    # #"invalid_order",
    # #"invalid_res_init1",
    # #"invalid_res_init2"
    # # ])
def test_invalid_res_shells_setting(tmp_environ,
                                    res_init, res_shells, error_message):
    if os.path.isdir("pairef_bad_res_shells_setting"):
        # Preparation - clean rests from previously examined test that
        # has been interrupted
        shutil.rmtree("pairef_bad_res_shells_setting")
    cp = run("--HKLIN " + str(config("lysozyme_l3600s_1-4A.mtz")) + ""
             " --XYZIN " + str(config("lysozyme_arp_model_2A.pdb")) + ""
             " -i " + res_init + ""
             " -r " + res_shells + ""
             " -p bad_res_shells_setting" + ""
             " -t")
    assert cp.returncode != 0
    # assert not cp.stdout
    assert cp.stderr == error_message


def test_valid_args(tmp_environ):
    if os.path.isdir("pairef_valid_args"):
        # Preparation - clean rests from previously examined test that
        # has been interrupted
        shutil.rmtree("pairef_valid_args")
    arguments = "--HKLIN " + str(config("lysozyme_l3600s_1-4A.mtz")) + \
        " --XYZIN " + str(config("lysozyme_arp_model_2A.pdb")) + \
        " -i 2" + \
        " -r " + RES_SHELLS_GOOD + \
        " -p valid_args" + \
        " -t"
    cp = run(arguments)
    expected_stdout_1 = """
 __   _  ___ __  ___ ___
 )_) /_)  )  )_) )_  )_
/   / / _(_ / \ (__ (

automatic PAIRed REFinement protocol
"""
# run date and time: ...
    expected_stdout_2 = """
Please cite: "Paired refinement under the control of PAIREF"
M. Maly, K. Diederichs, J. Dohnalek, P. Kolenko (2020) IUCrJ 7

Command line arguments: """ + arguments + """

Program has been executed with following input parameters:
"""
    expected_stdout_2 += " * XYZIN: " + str(os.path.dirname(__file__)) + "/" \
        "fixtures/lysozyme_arp_model_2A.pdb\n"
    expected_stdout_2 += " * HKLIN: " + str(os.path.dirname(__file__)) + "/" \
        "fixtures/lysozyme_l3600s_1-4A.mtz\n"
    expected_stdout_2 += """ * Project name: valid_args
 * Resolution shells: 1.9,1.8,1.7,1.6,1.5,1.4
 * Light-testing mode (REFMAC5 will not be executed).

"""
    expected_stdout_2 += "Resolution of the merged diffraction data " + \
        str(os.path.dirname(__file__)) + "/" \
        "fixtures/lysozyme_l3600s_1-4A.mtz: 39.35-1.40 A"
    expected_stdout_2 += """
Manual setting of initial high resolution limit will be used: 2.00 A.
High resolution diffraction limits: 1.90 A, 1.80 A, 1.70 A, 1.60 A, 1.50 A, 1.40 A
"""
    expected_stdout_2 += \
        " * Data with FreeRflag set 0 will be excluded during refinement.\n"
    expected_stdout_2 += "\n"
    expected_stdout_2 += "Current working directory: " \
        "" + tempfile.gettempdir() + "/pairef_valid_args\n"
    expected_stdout_2 += \
        "------> RESULTS AND THE CURRENT STATUS OF CALCULATIONS ARE LISTED" \
        " IN A HTML LOG FILE " + tempfile.gettempdir() + \
        "/pairef_valid_args/PAIREF_valid_args.html\n\n"
    stdout_all = cp.stdout.splitlines(True)
    stdout_1 = "".join(stdout_all[:6])
    stdout_2 = "".join(stdout_all[9:])
    assert stdout_1 == expected_stdout_1
    assert stdout_2 == expected_stdout_2
    assert os.path.isdir("pairef_valid_args")
    shutil.rmtree("pairef_valid_args")
    assert cp.returncode == 0
    assert not cp.stderr


def test_create_workdir(tmp_environ):
    # Preparation - clean rests from previously examined test that has been
    # interrupted
    if os.path.isdir("pairef_create_workdir"):
        os.rmdir("pairef_create_workdir")
    if os.path.isdir("pairef_create_workdir_new"):
        os.rmdir("pairef_create_workdir_new")
    # Test
    create_workdir("create_workdir")
    assert os.path.isdir("pairef_create_workdir")
    create_workdir("create_workdir")
    assert os.path.isdir("pairef_create_workdir_new")
    os.rmdir("pairef_create_workdir")
    os.rmdir("pairef_create_workdir_new")


def test_which():
    result = which("python")
    assert result
    result = which("ThisCommandShouldNotExist")
    assert not result
